#! /usr/bin/env python
"""Deduplication based on active learning (dedupe-io)"""
from __future__ import annotations

import logging
import os
import sqlite3
import statistics
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import dedupe as dedupe_io
import pandas as pd
import psutil
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from dedupe._typing import RecordDictPair as TrainingExample
from dedupe._typing import TrainingData
from dedupe.core import unique

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.dedupe.utils
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.dedupe

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-lines


@zope.interface.implementer(colrev.env.package_manager.DedupePackageEndpointInterface)
@dataclass
class ActiveLearningDedupeTraining(JsonSchemaMixin):
    """Active learning: training phase (minimum sample size of 50 required)"""

    # pylint: disable=too-many-instance-attributes

    __n_match: int
    __n_distinct: int
    __manual: bool

    settings_class = colrev.env.package_manager.DefaultSettings
    TRAINING_FILE_RELATIVE = Path(".records_dedupe_training.json")
    SETTINGS_FILE_RELATIVE = Path(".records_learned_settings")

    deduper: dedupe_io.Deduper

    ci_supported: bool = False

    # Code based on
    # https://github.com/dedupeio/dedupe-examples/blob/master/csv_example/csv_example.py

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,  # pylint: disable=unused-argument
        settings: dict,
    ):
        logging.basicConfig()
        logging.getLogger("dedupe.canopy_index").setLevel(logging.WARNING)

        self.settings = self.settings_class.load_settings(data=settings)
        self.training_file = (
            dedupe_operation.review_manager.path / self.TRAINING_FILE_RELATIVE
        )
        self.settings_file = (
            dedupe_operation.review_manager.path / self.SETTINGS_FILE_RELATIVE
        )

        self.local_index = dedupe_operation.review_manager.get_local_index()
        self.review_manager = dedupe_operation.review_manager

        if hasattr(dedupe_operation.review_manager, "dataset"):
            dedupe_operation.review_manager.dataset.update_gitignore(
                add=[self.SETTINGS_FILE_RELATIVE]
            )

    def setup_active_learning_dedupe(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        retrain: bool,
        in_memory: bool,
    ) -> None:
        """Prepare data for active learning setup"""
        # pylint: disable=import-outside-toplevel
        import random

        logging.getLogger("opensearch").setLevel(logging.ERROR)
        logging.getLogger("dedupe.training").setLevel(logging.WARNING)
        logging.getLogger("dedupe.api").setLevel(logging.WARNING)

        if retrain:
            # Note : removing the training_file would be to start from scratch...
            # self.training_file.unlink(missing_ok=True)
            self.settings_file.unlink(missing_ok=True)

        dedupe_operation.review_manager.logger.info("Import data ...")

        # Possible extension: in the read_data, we may want to append the colrev_status
        # to use Gazetteer (dedupe_io) if applicable (no duplicates in pos-md_processed)

        data_d = dedupe_operation.read_data()

        # to address memory issues, we select a sample from data_d
        # and feed it to prepare_training:
        # https://github.com/dedupeio/dedupe/issues/920

        if not in_memory:
            # Note: we have to make sure that when we sample for training,
            # the not-in-memory mode is used for duplicate clustering
            # otherwise, non-sampled duplicates will not be identified
            max_training_sample_size = min(1500, len(list(data_d.keys())))
            dedupe_operation.review_manager.logger.info(
                f"Selecting a random sample of {max_training_sample_size}"
                " to avoid memory problems"
            )
            # gh_issue https://github.com/CoLRev-Environment/colrev/issues/37
            # consider similar proportions of post-md_processed/md_prepared?
            keys = random.sample(list(data_d.keys()), max_training_sample_size)
            data_d = {key: data_d[key] for key in keys}

        # dedupe_operation.review_manager.logger.debug(
        #     dedupe_operation.review_manager.p_printer.pformat(data_d)
        # )

        # def title_corpus():
        #     for record in data_d.values():
        #         yield record["title"]

        # def container_corpus():
        #     for record in data_d.values():
        #         yield record["container_title"]

        # def author_corpus():
        #     for record in data_d.values():
        #         yield record["author"]

        # Training

        # gh_issue https://github.com/CoLRev-Environment/colrev/issues/37
        # creating a corpus from all fields may create memory issues...

        # Define the fields dedupe will pay attention to
        fields = [
            {
                "field": "author",
                "type": "String",
                # "corpus": author_corpus(),k
                "has missing": True,
                "crf": True,
            },
            {
                "field": "title",
                "type": "String",
                #  "corpus": title_corpus()
                "has missing": True,
                "crf": True,
            },
            {
                "field": "container_title",
                "variable name": "container_title",
                "type": "ShortString",
                # "corpus": container_corpus(),
                "has missing": True,
                "crf": True,
            },
            {
                "field": "year",
                "variable name": "year",
                "has missing": True,
                "type": "DateTime",
            },
            {
                "field": "volume",
                "variable name": "volume",
                "type": "ShortString",
                "has missing": True,
            },
            {
                "field": "number",
                "variable name": "number",
                "type": "ShortString",
                "has missing": True,
            },
            {
                "field": "pages",
                "type": "ShortString",
                "has missing": True,
                "crf": True,
            },
            {
                "type": "Interaction",
                "interaction variables": [
                    "container_title",
                    "year",
                    "volume",
                    "number",
                ],
            },
        ]
        # Consider using exists:
        # https://docs.dedupe.io/en/latest/Variable-definition.html#exists

        # Interactions:
        # https://docs.dedupe.io/en/latest/Variable-definition.html

        # Create a new deduper object and pass our data model to it.
        self.deduper = dedupe_io.Dedupe(fields)

        # If we have training data saved from a previous run of dedupe,
        # look for it and load it in.
        # __Note:__ if you want to train from scratch, delete the training_file

        if len(data_d) < 50:
            raise colrev_exceptions.DedupeError(
                "Sample size too small for active learning. "
                "Use simple_dedupe instead:\n"
                f"{colors.ORANGE}  colrev settings -m 'dedupe.dedupe_package_endpoints="
                f'[{{"endpoint":"colrev.simple_dedupe"}}]\'{colors.END}'
            )

        if self.training_file.is_file():
            dedupe_operation.review_manager.logger.info(
                "Reading pre-labeled training data from "
                f"{self.training_file.name} "
                "and preparing data"
            )
            dedupe_operation.review_manager.logger.info("Prepare training data")
            with open(self.training_file, "rb") as file:
                self.deduper.prepare_training(data_d, file)
        else:
            dedupe_operation.review_manager.logger.info("Prepare training data")
            # TBD: maybe use the sample_size parameter here?
            self.deduper.prepare_training(data_d)

        del data_d

        dedupe_operation.review_manager.logger.info(
            "Reading and preparation completed."
        )

    def __get_nr_duplicates(self, *, result_list: list) -> int:
        return len([i for i in result_list if "duplicate" == i["decision"]])

    def __get_nr_non_duplicates(self, *, result_list: list) -> int:
        return len([i for i in result_list if "no_duplicate" == i["decision"]])

    def apply_active_learning(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        results: list,
    ) -> None:
        """Apply the active learning results"""
        if (
            self.__get_nr_duplicates(result_list=results) > 10
            and self.__get_nr_non_duplicates(result_list=results) > 10
        ):
            dedupe_operation.apply_merges(results=results, complete_dedupe=False)

            dedupe_operation.review_manager.logger.info("Training deduper.")

            # Using the examples we just labeled, train the deduper and learn
            # blocking predicates
            self.deduper.train(recall=0.9, index_predicates=True)
            # print(self.deduper.data_model._field_comparators)
            # print(self.deduper.predicates)

            # When finished, save our training to disk
            with open(self.training_file, "w", encoding="utf-8") as train_file:
                self.deduper.write_training(train_file)
            dedupe_operation.review_manager.dataset.add_changes(path=self.training_file)

            # Save our weights and predicates to disk.  If the settings file
            # exists, we will skip all the training and learning next time we run
            # this file.
            with open(self.settings_file, "wb") as sett_file:
                self.deduper.write_settings(sett_file)
            # self.cleanup_training()

            dedupe_operation.review_manager.create_commit(
                msg="Labeling of duplicates (active learning)",
                manual_author=True,
            )

        else:
            dedupe_operation.review_manager.logger.info(
                "Not enough duplicates/non-duplicates to train deduper."
            )
            if self.__get_nr_duplicates(result_list=results) > 0:
                print([x for x in results if "duplicate" == x["decision"]])
            if self.__get_nr_non_duplicates(result_list=results) > 30:
                if "y" == input(
                    "Set remaining records to non-duplicated "
                    "(at least 50 non-duplicates recommended) (y,n)?"
                ):
                    dedupe_operation.apply_merges(results=results, complete_dedupe=True)
                    dedupe_operation.review_manager.create_commit(
                        msg="Set remaining records to non-duplicated (not enough to train ML)",
                        manual_author=True,
                    )
                    return

        if (
            self.__get_nr_duplicates(result_list=results) == 0
            and self.__get_nr_non_duplicates(result_list=results) > 100
        ):
            if input("Set remaining records to md_processed (no duplicates)?") == "y":
                dedupe_operation.apply_merges(results=results, complete_dedupe=True)
                dedupe_operation.review_manager.create_commit(
                    msg="Set remaining records to non-duplicated",
                    manual_author=True,
                )

    def __get_record_pair_for_labeling(
        self,
        *,
        use_previous: bool,
        examples_buffer: list,
        uncertain_pairs: list[TrainingExample],
    ) -> TrainingExample:
        if use_previous:
            record_pair, _ = examples_buffer.pop(0)
        else:
            # try:
            if not uncertain_pairs:
                uncertain_pairs = self.deduper.uncertain_pairs()

            record_pair = uncertain_pairs.pop()
            # except IndexError:
            #     break
        return record_pair

    def __identical_colrev_ids(self, *, record_pair: TrainingExample) -> bool:
        # if any of the colrev_ids NA,
        # we don't know whether we have a duplicate.
        return (
            record_pair[0]["colrev_id"] == record_pair[1]["colrev_id"]
            and "NA" != record_pair[0]["colrev_id"]
            and "NA" != record_pair[1]["colrev_id"]
        )

    def __active_label_duplicate(
        self,
        *,
        manual_dedupe_decision_list: typing.List[dict],
        record_pair: TrainingExample,
        examples_buffer: list[
            tuple[TrainingExample, typing.Literal["match", "distinct", "uncertain"]]
        ],
    ) -> None:
        manual_dedupe_decision_list.append(
            {
                "ID1": record_pair[0]["ID"],
                "ID2": record_pair[1]["ID"],
                "decision": "duplicate",
            }
        )
        examples_buffer.insert(0, (record_pair, "match"))
        msg = (
            f"Marked as duplicate: {record_pair[0]['ID']} - "
            + f"{record_pair[1]['ID']}"
        )
        self.review_manager.report_logger.info(msg)

    def __active_label_no_duplicate(
        self,
        *,
        manual_dedupe_decision_list: typing.List[dict],
        record_pair: TrainingExample,
        examples_buffer: list,
    ) -> None:
        if not self.__manual:
            # Ensure that non-dupes do not exceed 3x dupes
            # (for balanced training data)
            if self.__n_distinct > self.__n_match * 3:
                examples_buffer.insert(0, (record_pair, "uncertain"))
                # continue

        manual_dedupe_decision_list.append(
            {
                "ID1": record_pair[0]["ID"],
                "ID2": record_pair[1]["ID"],
                "decision": "no_duplicate",
            }
        )
        examples_buffer.insert(0, (record_pair, "distinct"))
        msg = (
            f"Marked as non-duplicate: {record_pair[0]['ID']}"
            + f" - {record_pair[1]['ID']}"
        )
        self.review_manager.report_logger.info(msg)

    def __active_label_uncertain(
        self, *, examples_buffer: list, record_pair: TrainingExample
    ) -> None:
        examples_buffer.insert(0, (record_pair, "uncertain"))

    def __active_label_check_finished(
        self, *, manual_dedupe_decision_list: list
    ) -> bool:
        nr_duplicates = self.__get_nr_duplicates(
            result_list=manual_dedupe_decision_list
        )
        nr_non_duplicates = self.__get_nr_non_duplicates(
            result_list=manual_dedupe_decision_list
        )
        if not nr_duplicates > 30 or not nr_non_duplicates > 30:
            if "y" != input(
                "The machine-learning requires "
                "30 duplicates and 30 non-duplicates. "
                "Quit anyway [y,n]?"
            ):
                return False
        os.system("cls" if os.name == "nt" else "clear")
        print("Finished labeling")
        return True

    def __check_finised_labeling(
        self, *, examples_buffer: list, max_associations_to_check: int
    ) -> bool:
        self.__n_match = len(self.deduper.training_pairs["match"]) + sum(
            label == "match" for _, label in examples_buffer
        )
        self.__n_distinct = len(self.deduper.training_pairs["distinct"]) + sum(
            label == "distinct" for _, label in examples_buffer
        )
        if (self.__n_match + self.__n_distinct) > max_associations_to_check:
            return True
        return False

    def __mark_active_labeling_pairs(self, *, examples_buffer: list) -> None:
        if len(examples_buffer) > 1:  # Max number of previous operations
            record_pair, label = examples_buffer.pop()
            if label in {"distinct", "match"}:
                examples: TrainingData = {"distinct": [], "match": []}
                examples[label].append(record_pair)
                self.deduper.mark_pairs(examples)

    def __adapted_console_label_active_label(
        self,
        *,
        manual_dedupe_decision_list: list,
        max_associations_to_check: int,
        keys: list,
        uncertain_pairs: list,
    ) -> list[tuple[TrainingExample, typing.Literal["match", "distinct", "uncertain"]]]:
        examples_buffer: list[
            tuple[TrainingExample, typing.Literal["match", "distinct", "uncertain"]]
        ] = []
        finished, use_previous = False, False

        while not finished:
            record_pair = self.__get_record_pair_for_labeling(
                use_previous=use_previous,
                examples_buffer=examples_buffer,
                uncertain_pairs=uncertain_pairs,
            )
            if use_previous:
                use_previous = False

            finished = self.__check_finised_labeling(
                examples_buffer=examples_buffer,
                max_associations_to_check=max_associations_to_check,
            )

            user_input = "u"
            if self.__identical_colrev_ids(record_pair=record_pair):
                user_input = "y"
            else:
                # Check local_index for duplicate information
                curations_dupe_info = self.local_index.is_duplicate(
                    record1_colrev_id=record_pair[0]["colrev_id"].split(";"),
                    record2_colrev_id=record_pair[1]["colrev_id"].split(";"),
                )

                user_input = (
                    colrev.ops.built_in.dedupe.utils.console_duplicate_instance_label(
                        record_pair,
                        keys,
                        self.__manual,
                        curations_dupe_info,
                        self.__n_match,
                        self.__n_distinct,
                        examples_buffer,
                    )
                )

            if user_input == "y":
                self.__active_label_duplicate(
                    manual_dedupe_decision_list=manual_dedupe_decision_list,
                    record_pair=record_pair,
                    examples_buffer=examples_buffer,
                )

            elif user_input == "n":
                self.__active_label_no_duplicate(
                    manual_dedupe_decision_list=manual_dedupe_decision_list,
                    record_pair=record_pair,
                    examples_buffer=examples_buffer,
                )

            elif user_input == "u":
                self.__active_label_uncertain(
                    examples_buffer=examples_buffer, record_pair=record_pair
                )

            elif user_input == "f":
                finished = self.__active_label_check_finished(
                    manual_dedupe_decision_list=manual_dedupe_decision_list
                )
                if not finished:
                    continue

            elif user_input == "p":
                use_previous = True
                uncertain_pairs.append(record_pair)

            self.__mark_active_labeling_pairs(examples_buffer=examples_buffer)

        return examples_buffer

    def __adapted_console_label(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        manual: bool,
        max_associations_to_check: int = 1000,
    ) -> None:
        """
        Train a matcher instance (Dedupe, RecordLink, or Gazetteer) from the cli.
        Example

        .. code:: python

        > deduper = dedupe.Dedupe(variables)
        > deduper.prepare_training(data)
        > dedupe.console_label(deduper)
        """

        dedupe_operation.review_manager.logger.info(
            "Note: duplicate associations available in the LocalIndex "
            "are applied automatically."
        )
        dedupe_operation.review_manager.logger.info("Press Enter to start.")
        input()
        self.__manual = manual

        keys = unique(
            field.field for field in self.deduper.data_model.primary_variables
        )

        uncertain_pairs: list[TrainingExample] = []
        manual_dedupe_decision_list: typing.List[dict] = []

        examples_buffer = self.__adapted_console_label_active_label(
            manual_dedupe_decision_list=manual_dedupe_decision_list,
            max_associations_to_check=max_associations_to_check,
            keys=keys,
            uncertain_pairs=uncertain_pairs,
        )

        for record_pair, label in examples_buffer:
            if label in ["distinct", "match"]:
                examples: typing.Dict[str, list] = {"distinct": [], "match": []}
                examples[label].append(record_pair)
                self.deduper.mark_pairs(examples)

        # Note : for debugging:
        # import csv
        # keys = manual_dedupe_decision_list[0].keys()
        # with open("manual_dedupe_decision_list.csv", "w", newline="") as output_file:
        #     dict_writer = csv.DictWriter(output_file, keys)
        #     dict_writer.writeheader()
        #     dict_writer.writerows(manual_dedupe_decision_list)

        # Apply and commit
        self.apply_active_learning(
            dedupe_operation=dedupe_operation,
            results=manual_dedupe_decision_list,
        )

    def run_dedupe(self, dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
        """Run the console labeling to train the active learning model"""

        # Setting in-memory mode depending on system RAM
        records_headers = dedupe_operation.review_manager.dataset.load_records_dict(
            header_only=True
        )
        sample_size = len(list(records_headers.values()))
        ram = psutil.virtual_memory().total
        in_memory = sample_size * 5000000 < ram

        self.setup_active_learning_dedupe(
            dedupe_operation=dedupe_operation, retrain=False, in_memory=in_memory
        )

        dedupe_io.console_label = self.__adapted_console_label
        dedupe_io.console_label(dedupe_operation=dedupe_operation, manual=True)


@zope.interface.implementer(colrev.env.package_manager.DedupePackageEndpointInterface)
@dataclass
class ActiveLearningDedupeAutomated(JsonSchemaMixin):
    """Applies trained (active learning) model"""

    settings: ActiveLearningSettings
    ci_supported: bool = False

    @dataclass
    class ActiveLearningSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for ActiveLearning"""

        endpoint: str
        merge_threshold: float = 0.8
        partition_threshold: float = 0.5

        _details = {
            "merge_threshold": {"tooltip": "Threshold for merging record pairs"},
            "partition_threshold": {"tooltip": "Threshold for partitioning"},
        }

    settings_class = ActiveLearningSettings

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,  # pylint: disable=unused-argument
        settings: dict,
    ):
        logging.basicConfig()
        logging.getLogger("dedupe.canopy_index").setLevel(logging.WARNING)

        self.settings = self.settings_class.load_settings(data=settings)

        self.settings_file = (
            dedupe_operation.review_manager.path
            / ActiveLearningDedupeTraining.SETTINGS_FILE_RELATIVE
        )

        assert self.settings.merge_threshold >= 0.0
        assert self.settings.merge_threshold <= 1.0
        assert self.settings.partition_threshold >= 0.0
        assert self.settings.partition_threshold <= 1.0

    def __get_duplicates_from_clusters(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, clustered_dupes: list
    ) -> list[dict]:
        dedupe_operation.review_manager.report_logger.info(
            f"set merge_threshold: {self.settings.merge_threshold}"
        )
        dedupe_operation.review_manager.logger.info(
            f"set merge_threshold: {self.settings.merge_threshold}"
        )
        results = []
        dedupe_decision_list = []
        for cluster_id, (records, scores) in enumerate(clustered_dupes):
            dedupe_decision_list.append(
                {
                    "cluster_id": cluster_id,
                    "records": list(records),
                    "score": statistics.mean(list(scores)),
                }
            )

        for dedupe_decision in dedupe_decision_list:
            if len(dedupe_decision["records"]) == 0:
                continue

            if dedupe_decision["score"] < self.settings.merge_threshold:
                continue

            orig_rec = dedupe_decision["records"].pop()
            if 0 == len(dedupe_decision["records"]):
                results.append(
                    {
                        "ID1": orig_rec,
                        "decision": "no_duplicate",
                    }
                )
                continue

            for dupe_rec in dedupe_decision["records"]:
                orig_propagated = dedupe_operation.review_manager.dataset.propagated_id(
                    record_id=orig_rec
                )
                dupe_propagated = dedupe_operation.review_manager.dataset.propagated_id(
                    record_id=dupe_rec
                )

                if not orig_propagated and not dupe_propagated:
                    # Use the record['ID'] without appended letters if possible
                    # Set orig_propagated=True if record_a_ID should be kept
                    if orig_rec[-1:].isnumeric() and not dupe_rec[-1:].isnumeric():
                        orig_propagated = True
                    else:
                        dupe_propagated = True
                        # This arbitrarily uses record_b_ID
                        # if none of the IDs has a letter appended.

                    if orig_propagated and dupe_propagated:
                        # both_IDs_propagated
                        dedupe_operation.review_manager.logger.error(
                            f"Both IDs propagated: {orig_rec}, {dupe_rec}"
                        )
                        continue

                    if orig_propagated:
                        results.append(
                            {
                                "ID1": orig_rec,
                                "ID2": dupe_rec,
                                "decision": "duplicate",
                                "score": dedupe_decision["score"],
                            }
                        )

                    else:
                        results.append(
                            {
                                "ID1": dupe_rec,
                                "ID2": orig_rec,
                                "decision": "duplicate",
                                "score": dedupe_decision["score"],
                            }
                        )
        return results

    def __highlight_cells(self, input_df: pd.DataFrame) -> pd.DataFrame:  # type: ignore  # noqa
        dataframe = input_df.copy()
        dataframe["cluster_id"] = dataframe["cluster_id"].astype(str)
        dataframe.loc[:, dataframe.columns != "cluster_id"] = "background-color: white"

        # http://www.excelsupersite.com/what-are-the-56-colorindex-colors-in-excel/
        available_colors = [
            "#FFFFFF",
            "#FFCC99",
            "#FFFFCC",
            "#CCFFCC",
            "#FFFF99",
            "#99CCFF",
            "#FF99CC",
        ]
        cur_color_index = -1
        cur_cluster = ""

        prev_row: dict = {}
        for i, row in dataframe.iterrows():
            if row["cluster_id"] != cur_cluster:
                cur_color_index += 1
                cur_cluster = row["cluster_id"]
            # dataframe.at[i, 'cluster_id'] = ( # only the cluster_id column
            dataframe.at[i, :] = (
                "background-color: "
                + available_colors[cur_color_index % len(available_colors)]
            )

        for i, row in input_df.iterrows():
            if i in [0, 1]:
                continue
            if len(prev_row) != 0:
                for j, val in row.items():
                    # changes in these fields should not be marked
                    if j in ["error", "confidence_score", "ID"]:
                        continue
                    # do not mark changes between different clusters
                    if j == "cluster_id" and prev_row["cluster_id"] != val:
                        break
                    if val != prev_row[j]:
                        dataframe.at[i, j] = dataframe.at[i, j] + "; font-weight: bold"

            prev_row = row.to_dict()

        return dataframe

    def __export_duplicates_excel(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, collected_duplicates: list
    ) -> None:
        if len(collected_duplicates) == 0:
            print("No duplicates found")
            return

        duplicates_df = pd.DataFrame.from_records(collected_duplicates)
        duplicates_df.fillna("", inplace=True)
        duplicates_df["distinct_str"] = (
            duplicates_df["author"]
            + duplicates_df["title"]
            + duplicates_df["year"]
            + duplicates_df["container_title"]
            + duplicates_df["volume"]
            + duplicates_df["number"]
            + duplicates_df["pages"]
        )
        # Only export bibliographically distict cases
        duplicates_df = duplicates_df.groupby("distinct_str").filter(
            lambda x: len(x) == 1
        )
        duplicates_df.drop(columns=["distinct_str"], inplace=True)

        duplicates_df = duplicates_df[
            [
                "error",
                "confidence_score",
                "cluster_id",
                "ID",
                "author",
                "title",
                "year",
                "container_title",
                "volume",
                "number",
                "pages",
            ]
        ]

        duplicates_df = duplicates_df.groupby("cluster_id").filter(lambda x: len(x) > 1)
        duplicates_df = duplicates_df.sort_values(
            ["confidence_score", "cluster_id"], ascending=(True, False)
        )
        duplicates_df["confidence_score"] = duplicates_df["confidence_score"].round(4)
        # to adjust column widths in ExcelWriter:
        # http://pandas-docs.github.io/pandas-docs-travis/user_guide/style.html
        duplicates_df.style.apply(self.__highlight_cells, axis=None)
        duplicates_df.to_excel(dedupe_operation.dupe_file, index=False)

    def __export_non_duplicates_excel(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        collected_non_duplicates: list,
    ) -> None:
        if len(collected_non_duplicates) == 0:
            print("No duplicates.")
            return

        non_duplicates_df = pd.DataFrame.from_records(collected_non_duplicates)
        # To develop in jupyter:
        # non_duplicates_df.to_csv(output_file, index=False)
        # non_duplicates_df = pd.read_csv("duplicates_for_validation.csv")
        non_duplicates_df = non_duplicates_df[
            [
                "error",
                "cluster_id",
                "confidence_score",
                "ID",
                "author",
                "title",
                "year",
                "container_title",
                "volume",
                "number",
                "pages",
            ]
        ]
        non_duplicates_df = non_duplicates_df.groupby("cluster_id").filter(
            lambda x: len(x) > 1
        )
        non_duplicates_df = non_duplicates_df.sort_values(
            ["confidence_score", "cluster_id"], ascending=(True, False)
        )
        non_duplicates_df["confidence_score"] = non_duplicates_df[
            "confidence_score"
        ].round(4)
        # to adjust column widths in ExcelWriter:
        # http://pandas-docs.github.io/pandas-docs-travis/user_guide/style.html
        non_duplicates_df.style.apply(self.__highlight_cells, axis=None)
        non_duplicates_df.to_excel(dedupe_operation.non_dupe_file_xlsx, index=False)

    def __get_collected_dupes_non_dupes_from_clusters(
        self, *, clustered_dupes: list, data_d: dict
    ) -> dict:
        # pylint: disable=too-many-locals

        cluster_membership = {}
        # cluster_membership:
        # {'FrolovaFrolovKayurovEtAl2021': {'Cluster ID': 352, 'confidence_score': 1.0},
        #  'BhaskaraBawa2021': {'Cluster ID': 353, 'confidence_score': 1.0}}
        for cluster_id, (records, scores) in enumerate(clustered_dupes):
            for record_id, score in zip(records, scores):
                cluster_membership[record_id] = {
                    "cluster_id": cluster_id,
                    "confidence_score": score,
                }

        results: typing.Dict[str, list] = {
            "collected_duplicates": [],
            "collected_non_duplicates": [],
        }
        for cluster_id, vals in data_d.items():
            vals.update(error="")
            if cluster_id in cluster_membership:
                cur_cluster_membership = cluster_membership[cluster_id]
                vals.update(cur_cluster_membership)
                if (
                    cur_cluster_membership["confidence_score"]
                    > self.settings.merge_threshold
                ):
                    results["collected_duplicates"].append(vals)
                else:
                    results["collected_non_duplicates"].append(vals)

        # Set confidence scores to average of group
        for cluster_nr in {d["cluster_id"] for d in results["collected_duplicates"]}:
            avg_confidence = statistics.mean(
                [
                    d["confidence_score"]
                    for d in results["collected_duplicates"]
                    if d["cluster_id"] == cluster_nr
                ]
            )
            for collected_duplicate in results["collected_duplicates"]:
                if collected_duplicate["cluster_id"] == cluster_nr:
                    collected_duplicate["confidence_score"] = avg_confidence
        for cluster_nr in {
            d["cluster_id"] for d in results["collected_non_duplicates"]
        }:
            avg_confidence = statistics.mean(
                [
                    d["confidence_score"]
                    for d in results["collected_non_duplicates"]
                    if d["cluster_id"] == cluster_nr
                ]
            )
            for collected_non_duplicate in results["collected_non_duplicates"]:
                if collected_non_duplicate["cluster_id"] == cluster_nr:
                    collected_non_duplicate["confidence_score"] = avg_confidence

        return results

    def __export_validation_excel(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        clustered_dupes: list,
        data_d: dict,
    ) -> None:
        results = self.__get_collected_dupes_non_dupes_from_clusters(
            clustered_dupes=clustered_dupes, data_d=data_d
        )

        self.__export_duplicates_excel(
            dedupe_operation=dedupe_operation,
            collected_duplicates=results["collected_duplicates"],
        )

        self.__export_non_duplicates_excel(
            dedupe_operation=dedupe_operation,
            collected_non_duplicates=results["collected_non_duplicates"],
        )

    def __cluster_duplicates(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, data_d: dict
    ) -> list:
        # pylint: disable=too-many-locals

        dedupe_operation.review_manager.logger.info("Clustering duplicates...")
        dedupe_operation.review_manager.logger.info(
            f"Number of records (before): {len(data_d.items())}"
        )

        # Setting in-memory mode depending on system RAM
        records_headers = dedupe_operation.review_manager.dataset.load_records_dict(
            header_only=True
        )
        sample_size = len(list(records_headers.values()))
        ram = psutil.virtual_memory().total
        in_memory = sample_size * 5000000 < ram

        with open(self.settings_file, "rb") as sett_file:
            deduper = dedupe_io.StaticDedupe(sett_file, num_cores=4)

        # `partition` will return sets of records that dedupe
        # believes are all referring to the same entity.

        if in_memory:
            dedupe_operation.review_manager.report_logger.info(
                f"set partition_threshold: {self.settings.partition_threshold}"
            )

            clustered_dupes = deduper.partition(
                data_d, self.settings.partition_threshold
            )

            # from dedupe.core import BlockingError
            # except BlockingError:

            #     dedupe_operation.review_manager.logger.info(
            #         "No duplicates found (please check carefully)"
            #     )
            #     dedupe_operation.apply_merges(results=[], complete_dedupe=True)
            #     dedupe_operation.review_manager.create_commit(
            #         msg="Merge duplicate records (no duplicates detected)",
            #     )
            #     dedupe_operation.review_manager.logger.info(
            #         "If there are errors, it could be necessary to remove the "
            #         ".records_dedupe_training.json to train a fresh dedupe model."
            #     )

            #     pass
            # except KeyboardInterrupt:
            #     print("KeyboardInterrupt")
            #     pass

        else:
            for field in deduper.fingerprinter.index_fields:
                field_data = (r[field] for r in data_d.values() if field in r)
                deduper.fingerprinter.index(field_data, field)

            full_data = ((r["ID"], r) for r in data_d.values())

            # pylint: disable=not-callable
            # fingerprinter is callable according to
            # https://github.com/dedupeio/dedupe/blob/
            # b9d8f111bcd5ffd177659f79f57354d9a9318359/dedupe/blocking.py
            b_data = deduper.fingerprinter(full_data)

            # use sqlite: light-weight, file-based
            # https://docs.python.org/3/library/sqlite3.html
            # https://dedupeio.github.io/dedupe-examples/docs/pgsql_big_dedupe_example.html

            dedupe_db = Path("dedupe.db")
            dedupe_db.unlink(missing_ok=True)
            con = sqlite3.connect(str(dedupe_db))

            cur = con.cursor()

            cur.execute("""DROP TABLE IF EXISTS blocking_map""")
            cur.execute("""CREATE TABLE blocking_map (block_key text, ID INTEGER)""")
            cur.executemany("""INSERT into blocking_map values (?, ?)""", b_data)

            records_data = {r["ID"]: r for r in data_d.values()}

            def record_pairs(result_set: list[tuple]) -> typing.Iterator[tuple]:
                for row in result_set:
                    id_a, id_b = row
                    record_a = (id_a, records_data[id_a])
                    record_b = (id_b, records_data[id_b])

                    yield record_a, record_b

            cur.execute(
                """select DISTINCT l.ID as east, r.ID as west
                        from blocking_map as l
                        INNER JOIN blocking_map as r
                        using (block_key)
                        where east != west"""
            )

            clustered_dupes = list(
                deduper.cluster(
                    deduper.score(record_pairs(cur.fetchall())), threshold=0.5
                )
            )

            # import csv
            # clusterin_results_csv = Path("clusterin_results.csv")
            # clusterin_results_csv.unlink(missing_ok=True)
            # with open(clusterin_results_csv, "w") as out:
            #     csv_out = csv.writer(out)
            #     csv_out.writerow(["ID1", "ID2", "conf"])
            #     for row in list(cluster_ids(clustered_dupes)):
            #         if row[0] != row[1]:  # only focus on non-identical IDs
            #             csv_out.writerow(row)

            con.commit()
            con.close()
            dedupe_db.unlink(missing_ok=True)

        dedupe_operation.review_manager.report_logger.info(
            f"Number of duplicate sets {len(clustered_dupes)}"
        )
        return clustered_dupes

    def run_dedupe(self, dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
        """Cluster potential duplicates, merge, and export validation tables"""

        if not self.settings_file.is_file():
            dedupe_operation.review_manager.logger.info(
                "No settings file. Skip ML-clustering."
            )
            return

        data_d = dedupe_operation.read_data()

        clustered_dupes = self.__cluster_duplicates(
            dedupe_operation=dedupe_operation, data_d=data_d
        )

        results = self.__get_duplicates_from_clusters(
            dedupe_operation=dedupe_operation, clustered_dupes=clustered_dupes
        )

        dedupe_operation.apply_merges(results=results, complete_dedupe=True)

        self.__export_validation_excel(
            dedupe_operation=dedupe_operation,
            clustered_dupes=clustered_dupes,
            data_d=data_d,
        )

        dedupe_operation.review_manager.create_commit(
            msg="Merge duplicate records (based on active-learning clusters)",
            script_call="colrev dedupe",
        )

        dedupe_operation.review_manager.logger.info(
            "Successfully completed the deduplication. Please check the "
            "duplicates_to_validate.xlsx and non_duplicates_to_validate.xlsx for "
            'potential errors.\nTo fix them, mark them in the "error" column and '
            "run\n  colrev dedupe --fix_errors\n\n"
        )

        if Path("same_source_merges.txt").is_file():
            dedupe_operation.review_manager.logger.info(
                "Detected and prevented same-source merges. Please check potential"
                "duplicates in same_source_merges.txt"
            )

        info = dedupe_operation.get_info()
        if len(info["same_source_merges"]) > 0:
            dedupe_operation.review_manager.logger.info(
                f"\n{colors.ORANGE}Same source merges to check:{colors.END}"
                "\n- ".join(info["same_source_merges"]) + "\n"
            )
        else:
            dedupe_operation.review_manager.logger.info(
                "\nNo same-origin merges detected."
            )



#! /usr/bin/env python
"""Source-specific preparation as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class SourceSpecificPrep(JsonSchemaMixin):
    """Prepares records based on the prepare scripts specified by the SearchSource"""

    source_correction_hint = "check with the developer"
    ci_supported: bool = True

    always_apply_changes = True
    settings_class = colrev.env.package_manager.DefaultSettings

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        self.search_sources = colrev.ops.search_sources.SearchSources(
            review_manager=prep_operation.review_manager
        )

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by applying source-specific fixes"""

        # Note : we take the first origin (ie., the source-specific prep should
        # be one of the first in the prep-list)
        origin_source = record.data["colrev_origin"][0].split("/")[0]

        sources = [
            s
            for s in prep_operation.review_manager.settings.sources
            if s.filename == Path("data/search") / Path(origin_source)
        ]

        for source in sources:
            if source.endpoint not in self.search_sources.packages:
                continue
            endpoint = self.search_sources.packages[source.endpoint]

            if callable(endpoint.prepare):
                record = endpoint.prepare(record, source)
            else:
                print(f"error: {source.endpoint}")

        if "howpublished" in record.data and "url" not in record.data:
            if "url" in record.data["howpublished"]:
                record.rename_field(key="howpublished", new_key="url")
                record.update_field(
                    key="url",
                    value=record.data["url"].replace("\\url{", "").rstrip("}"),
                    source="source_specific_prep",
                )

        if "webpage" == record.data["ENTRYTYPE"].lower() or (
            "misc" == record.data["ENTRYTYPE"].lower() and "url" in record.data
        ):
            record.update_field(
                key="ENTRYTYPE", value="online", source="source_specific_prep"
            )

        return record



#! /usr/bin/env python
"""Synchronize records into a non-CoLRev project."""
import logging
import re
import typing
from pathlib import Path

import pybtex.errors
from pybtex.database.input import bibtex

import colrev.dataset
import colrev.env.local_index
import colrev.record
import colrev.ui_cli.cli_colors as colors


class Sync:
    """Synchronize records into a non-CoLRev repository"""

    cited_papers: list

    def __init__(self) -> None:
        self.records_to_import: typing.List[colrev.record.Record] = []
        self.non_unique_for_import: typing.List[typing.Dict] = []

        self.logger = self.__setup_logger(level=logging.DEBUG)
        self.paper_md = self.__get_md_file()

    def __setup_logger(self, *, level: int = logging.INFO) -> logging.Logger:
        """Setup the sync logger"""
        # pylint: disable=duplicate-code

        # for logger debugging:
        # from logging_tree import printout
        # printout()
        logger = logging.getLogger(f"colrev{str(Path.cwd()).replace('/', '_')}")
        logger.setLevel(level)

        if logger.handlers:
            for handler in logger.handlers:
                logger.removeHandler(handler)

        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        handler.setLevel(level)

        logger.addHandler(handler)
        logger.propagate = False

        return logger

    def __get_md_file(self) -> Path:
        paper_md = Path("")
        if Path("paper.md").is_file():
            paper_md = Path("paper.md")
        if Path("data/paper.md").is_file():
            paper_md = Path("data/paper.md")
        elif Path("review.md").is_file():
            paper_md = Path("review.md")
        return paper_md

    def __get_cited_papers_citation_keys(self) -> list:
        rst_files = list(Path.cwd().rglob("*.rst"))

        citation_keys = []
        if self.paper_md.is_file():
            self.logger.info("Load cited references from paper.md")
            content = self.paper_md.read_text(encoding="utf-8")
            res = re.findall(r"(^|\s|\[|;)(@[a-zA-Z0-9_]+)+", content)
            citation_keys = list({r[1].replace("@", "") for r in res})
            if "fig" in citation_keys:
                citation_keys.remove("fig")
            self.logger.info("Citations in paper.md: %s", len(citation_keys))
            self.cited_papers = citation_keys

        elif len(rst_files) > 0:
            self.logger.info("Load cited references from *.rst")
            for rst_file in rst_files:
                content = rst_file.read_text()
                res = re.findall(r":cite:p:`(.*)`", content)
                cited = [c for cit_group in res for c in cit_group.split(",")]
                citation_keys.extend(cited)

            citation_keys = list(set(citation_keys))
            self.logger.info("Citations in *.rst: %s", len(citation_keys))
            self.cited_papers = citation_keys

        else:
            print("Not found paper.md or *.rst")
        return citation_keys

    def get_cited_papers_from_source(self, *, src: Path) -> None:
        """Get the cited papers from a source file"""

        citation_keys = self.__get_cited_papers_citation_keys()

        ids_in_bib = self.__get_ids_in_paper()
        self.logger.info("References in bib: %s", len(ids_in_bib))

        if src.suffix == ".bib":
            parser = bibtex.Parser()
            bib_data = parser.parse_file(str(src))
            refs_in_src = colrev.dataset.Dataset.parse_records_dict(
                records_dict=bib_data.entries
            )
        else:
            print("Format not supported")
            return

        for citation_key in citation_keys:
            if citation_key in ids_in_bib:
                continue
            if citation_key not in refs_in_src:
                print(f"{citation_key} not in {src}")
                continue
            self.records_to_import.append(
                colrev.record.Record(data=refs_in_src[citation_key])
            )

    def get_cited_papers(self) -> None:
        """Get the cited papers"""

        citation_keys = self.__get_cited_papers_citation_keys()

        ids_in_bib = self.__get_ids_in_paper()
        self.logger.info("References in bib: %s", len(ids_in_bib))

        local_index = colrev.env.local_index.LocalIndex()
        for citation_key in citation_keys:
            if citation_key in ids_in_bib:
                continue

            if Path(f"{citation_key}.pdf").is_file():
                print("TODO - prefer!")
                # continue if found/extracted

            returned_records = local_index.search(
                query=f"citation_key='{citation_key}'"
            )

            if 0 == len(returned_records):
                self.logger.info("Not found: %s", citation_key)
            elif 1 == len(returned_records):
                if returned_records[0].data["ID"] in [
                    r.data["ID"] for r in self.records_to_import
                ]:
                    continue
                self.records_to_import.append(returned_records[0])
            else:
                listed_item: typing.Dict[str, typing.List] = {citation_key: []}
                for returned_record in returned_records:  # type: ignore
                    listed_item[citation_key].append(returned_record)
                self.non_unique_for_import.append(listed_item)

    def __get_ids_in_paper(self) -> typing.List:
        pybtex.errors.set_strict_mode(False)

        if Path("references.bib").is_file():
            parser = bibtex.Parser()
            bib_data = parser.parse_file(str(Path("references.bib")))
            records = colrev.dataset.Dataset.parse_records_dict(
                records_dict=bib_data.entries
            )

        else:
            records = {}

        return list(records.keys())

    def add_to_records_to_import(self, *, record: colrev.record.Record) -> None:
        """Add a record to the records_to_import list"""
        if record.data["ID"] not in [r.data["ID"] for r in self.records_to_import]:
            self.records_to_import.append(record)

    def __save_to_bib(self, *, records: dict, save_path: Path) -> None:
        # pylint: disable=duplicate-code

        def parse_bibtex_str(*, recs_dict_in: dict) -> str:
            def format_field(field: str, value: str) -> str:
                padd = " " * max(0, 28 - len(field))
                return f",\n   {field} {padd} = {{{value}}}"

            bibtex_str = ""

            first = True
            for record_id, record_dict in recs_dict_in.items():
                if not first:
                    bibtex_str += "\n"
                first = False

                bibtex_str += f"@{record_dict['ENTRYTYPE']}{{{record_id}"

                field_order = [
                    "doi",
                    "colrev.dblp.dblp_key",
                    "colrev.semantic_scholar.id",
                    "colrev.web_of_science.unique-id",
                    "author",
                    "booktitle",
                    "journal",
                    "title",
                    "year",
                    "volume",
                    "number",
                    "pages",
                    "editor",
                ]

                for ordered_field in field_order:
                    if ordered_field in record_dict:
                        if record_dict[ordered_field] == "":
                            continue
                        if isinstance(record_dict[ordered_field], (list, dict)):
                            continue
                        bibtex_str += format_field(
                            ordered_field, record_dict[ordered_field]
                        )

                for key, value in record_dict.items():
                    if key in field_order + ["ID", "ENTRYTYPE"]:
                        continue
                    if isinstance(key, (list, dict)):
                        continue

                    bibtex_str += format_field(key, value)

                bibtex_str += ",\n}\n"

            return bibtex_str

        bibtex_str = parse_bibtex_str(recs_dict_in=records)

        with open(save_path, "w", encoding="utf-8") as out:
            out.write(bibtex_str)

    def add_to_bib(self) -> None:
        """Add records to the bibliography"""

        if not self.paper_md.is_file():
            return

        if self.paper_md.read_text(encoding="utf-8").startswith("---"):
            self.__export_to_bib()

        else:
            # Append to # References if no header (mardown with linked)
            self.__append_as_citations()

    def __append_as_citations(self) -> None:
        if "# References" in self.paper_md.read_text(encoding="utf-8"):
            print("Already contains a reference section.")
            return

        with open(self.paper_md, "a", encoding="utf-8") as file:
            file.write("\n# References\n\n")
            ref_list = [
                record_to_import.format_bib_style()
                for record_to_import in self.records_to_import
            ]
            file.write("\n".join(sorted(ref_list)))

    def __export_to_bib(self) -> None:
        pybtex.errors.set_strict_mode(False)

        references_file = Path("references.bib")
        if not references_file.is_file():
            records = []
        else:
            parser = bibtex.Parser()
            bib_data = parser.parse_file(str(references_file))
            records = list(
                colrev.dataset.Dataset.parse_records_dict(
                    records_dict=bib_data.entries
                ).values()
            )

        available_ids = [r["ID"] for r in records]
        added = []
        for record_to_import in self.records_to_import:
            if record_to_import.data["ID"] not in available_ids:
                record_to_import = colrev.record.Record(
                    data={
                        k: v
                        for k, v in record_to_import.data.items()
                        if k not in colrev.record.Record.provenance_keys
                    }
                )
                records.append(record_to_import.get_data())
                available_ids.append(record_to_import.data["ID"])
                added.append(record_to_import)

        if len(added) > 0:
            self.logger.info("Loaded:")
            print()
            for added_record in added:
                added_record.print_citation_format()
            print()

            self.logger.info(
                "%s Loaded %s papers%s", colors.GREEN, len(added), colors.END
            )

        records_dict = {r["ID"]: r for r in records if r["ID"] in self.cited_papers}

        self.__save_to_bib(records=records_dict, save_path=references_file)



#! /usr/bin/env python
"""SearchSource: Crossref"""
from __future__ import annotations

import json
import re
import typing
import urllib
from copy import deepcopy
from dataclasses import dataclass
from importlib.metadata import version
from multiprocessing import Lock
from pathlib import Path
from sqlite3 import OperationalError
from typing import Optional
from typing import TYPE_CHECKING

import requests
import zope.interface
from crossref.restful import Etiquette
from crossref.restful import Journals
from crossref.restful import Works
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from thefuzz import fuzz

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.doi_org as doi_connector
import colrev.ops.built_in.search_sources.utils as connector_utils
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.search
    import colrev.ops.prep

# pylint: disable=unused-argument
# pylint: disable=duplicate-code
# pylint: disable=too-many-lines


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class CrossrefSearchSource(JsonSchemaMixin):
    """SearchSource for the Crossref API"""

    __ISSN_REGEX = r"^\d{4}-?\d{3}[\dxX]$"
    __YEAR_SCOPE_REGEX = r"^\d{4}-\d{4}$"

    # https://github.com/CrossRef/rest-api-doc
    __api_url = "https://api.crossref.org/works?"

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "doi"
    # "https://api.crossref.org/works/{{doi}}"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.oni
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/crossref.md"
    )
    short_name = "Crossref"
    __crossref_md_filename = Path("data/search/md_crossref.bib")

    __availability_exception_message = (
        f"Crossref ({colors.ORANGE}check https://status.crossref.org/{colors.END})"
    )

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        if settings:
            # Crossref as a search_source
            self.search_source = from_dict(
                data_class=self.settings_class, data=settings
            )
        else:
            # Crossref as an md-prep source
            crossref_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self.__crossref_md_filename
            ]
            if crossref_md_source_l:
                self.search_source = crossref_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint="colrev.crossref",
                    filename=self.__crossref_md_filename,
                    search_type=colrev.settings.SearchType.OTHER,
                    search_parameters={},
                    comment="",
                )

            self.crossref_lock = Lock()

        self.language_service = colrev.env.language_service.LanguageService()

        self.review_manager = source_operation.review_manager
        self.etiquette = self.get_etiquette(review_manager=self.review_manager)
        self.email = self.review_manager.get_committer()

    @classmethod
    def get_etiquette(
        cls, *, review_manager: colrev.review_manager.ReviewManager
    ) -> Etiquette:
        """Get the etiquette for the crossref api"""
        _, email = review_manager.get_committer()
        return Etiquette(
            "CoLRev",
            version("colrev"),
            "https://github.com/CoLRev-Environment/colrev",
            email,
        )

    def check_availability(
        self, *, source_operation: colrev.operation.Operation
    ) -> None:
        """Check status (availability) of the Crossref API"""

        try:
            # pylint: disable=duplicate-code
            test_rec = {
                "doi": "10.17705/1cais.04607",
                "author": "Schryen, Guido and Wagner, Gerit and Benlian, Alexander "
                "and Paré, Guy",
                "title": "A Knowledge Development Perspective on Literature Reviews: "
                "Validation of a new Typology in the IS Field",
                "ID": "SchryenEtAl2021",
                "journal": "Communications of the Association for Information Systems",
                "ENTRYTYPE": "article",
            }
            returned_record = self.crossref_query(
                record_input=colrev.record.PrepRecord(data=test_rec),
                jour_vol_iss_list=False,
                timeout=20,
            )[0]

            if 0 != len(returned_record.data):
                assert returned_record.data["title"] == test_rec["title"]
                assert returned_record.data["author"] == test_rec["author"]
            else:
                if not source_operation.force_mode:
                    raise colrev_exceptions.ServiceNotAvailableException(
                        self.__availability_exception_message
                    )
        except (requests.exceptions.RequestException, IndexError) as exc:
            print(exc)
            if not source_operation.force_mode:
                raise colrev_exceptions.ServiceNotAvailableException(
                    self.__availability_exception_message
                ) from exc

    def __query(self, **kwargs) -> typing.Iterator[dict]:  # type: ignore
        """Get records from Crossref based on a bibliographic query"""

        works = Works(etiquette=self.etiquette)
        # use facets:
        # https://api.crossref.org/swagger-ui/index.html#/Works/get_works

        crossref_query_return = works.query(**kwargs).sort("deposited").order("desc")
        yield from crossref_query_return

    @classmethod
    def query_doi(cls, *, doi: str, etiquette: Etiquette) -> colrev.record.PrepRecord:
        """Get records from Crossref based on a doi query"""

        try:
            works = Works(etiquette=etiquette)
            crossref_query_return = works.doi(doi)
            if crossref_query_return is None:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="Record not found in crossref (based on doi)"
                )

            retrieved_record_dict = connector_utils.json_to_record(
                item=crossref_query_return
            )
            retrieved_record = colrev.record.PrepRecord(data=retrieved_record_dict)
            return retrieved_record

        except (requests.exceptions.RequestException,) as exc:
            raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                msg="Record not found in crossref (based on doi)"
            ) from exc

    def __query_journal(self, *, issn: str, rerun: bool) -> typing.Iterator[dict]:
        """Get records of a selected journal from Crossref"""

        assert re.match(self.__ISSN_REGEX, issn)

        journals = Journals(etiquette=self.etiquette)
        if rerun:
            # Note : the "deposited" field is not always provided.
            # only the general query will return all records.
            crossref_query_return = journals.works(issn).query()
        else:
            crossref_query_return = (
                journals.works(issn).query().sort("deposited").order("desc")
            )

        yield from crossref_query_return

    def __create_query_url(
        self, *, record: colrev.record.Record, jour_vol_iss_list: bool
    ) -> str:
        if jour_vol_iss_list:
            if not all(x in record.data for x in ["journal", "volume", "number"]):
                raise colrev_exceptions.NotEnoughDataToIdentifyException
            params = {"rows": "50"}
            container_title = re.sub(r"[\W]+", " ", record.data["journal"])
            params["query.container-title"] = container_title.replace("_", " ")

            query_field = ""
            if "volume" in record.data:
                query_field = record.data["volume"]
            if "number" in record.data:
                query_field = query_field + "+" + record.data["number"]
            params["query"] = query_field

        else:
            if "title" not in record.data:
                raise colrev_exceptions.NotEnoughDataToIdentifyException()

            params = {"rows": "15"}
            if not isinstance(record.data.get("year", ""), str) or not isinstance(
                record.data.get("title", ""), str
            ):
                print("year or title field not a string")
                print(record.data)
                raise AssertionError

            bibl = (
                record.data["title"].replace("-", "_")
                + " "
                + record.data.get("year", "")
            )
            bibl = re.sub(r"[\W]+", "", bibl.replace(" ", "_"))
            params["query.bibliographic"] = bibl.replace("_", " ")

            container_title = record.get_container_title()
            if "." not in container_title:
                container_title = container_title.replace(" ", "_")
                container_title = re.sub(r"[\W]+", "", container_title)
                params["query.container-title"] = container_title.replace("_", " ")

            author_last_names = [
                x.split(",")[0] for x in record.data.get("author", "").split(" and ")
            ]
            author_string = " ".join(author_last_names)
            author_string = re.sub(r"[\W]+", "", author_string.replace(" ", "_"))
            params["query.author"] = author_string.replace("_", " ")

        url = self.__api_url + urllib.parse.urlencode(params)
        return url

    def __get_similarity(
        self, *, record: colrev.record.Record, retrieved_record_dict: dict
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record_dict.get("title", "NA").lower(),
            record.data.get("title", "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            colrev.record.PrepRecord(data=retrieved_record_dict)
            .get_container_title()
            .lower(),
            record.get_container_title().lower(),
        )
        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]

        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        # logger.debug(f'record: {pp.pformat(record)}')
        # logger.debug(f'similarities: {similarities}')
        # logger.debug(f'similarity: {similarity}')
        # pp.pprint(retrieved_record_dict)
        return similarity

    def __prep_crossref_record(
        self,
        *,
        record: colrev.record.Record,
        prep_main_record: bool = True,
        crossref_source: str = "",
    ) -> None:
        if "language" in record.data:
            try:
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data["language"]

        doi_connector.DOIConnector.get_link_from_doi(
            review_manager=self.review_manager,
            record=record,
        )

        if (
            self.review_manager.settings.is_curated_masterdata_repo()
        ) and "cited_by" in record.data:
            del record.data["cited_by"]

        if not prep_main_record:
            # Skip steps for feed records
            return

        if "retracted" in record.data.get(
            "warning", ""
        ) or "retracted" in record.data.get("prescreen_exclusion", ""):
            record.prescreen_exclude(reason="retracted")
            record.remove_field(key="warning")
        else:
            assert "" != crossref_source
            record.set_masterdata_complete(
                source=crossref_source,
                masterdata_repository=self.review_manager.settings.is_curated_repo(),
            )
            record.set_status(target_state=colrev.record.RecordState.md_prepared)

    def __get_crossref_query_items(
        self, *, record: colrev.record.Record, jour_vol_iss_list: bool, timeout: int
    ) -> list:
        # Note : only returning a multiple-item list for jour_vol_iss_list
        try:
            url = self.__create_query_url(
                record=record, jour_vol_iss_list=jour_vol_iss_list
            )
            headers = {"user-agent": f"{__name__} (mailto:{self.email})"}
            session = self.review_manager.get_cached_session()

            # review_manager.logger.debug(url)
            ret = session.request("GET", url, headers=headers, timeout=timeout)
            ret.raise_for_status()
            if ret.status_code != 200:
                # review_manager.logger.debug(
                #     f"crossref_query failed with status {ret.status_code}"
                # )
                return []

            data = json.loads(ret.text)

        # pylint: disable=duplicate-code
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc
        except (
            colrev_exceptions.NotEnoughDataToIdentifyException,
            json.decoder.JSONDecodeError,
            requests.exceptions.RequestException,
        ):
            return []

        return data["message"].get("items", [])

    def crossref_query(
        self,
        *,
        record_input: colrev.record.Record,
        jour_vol_iss_list: bool = False,
        timeout: int = 40,
    ) -> list:
        """Retrieve records from Crossref based on a query"""

        record = record_input.copy_prep_rec()
        record_list, most_similar, most_similar_record = [], 0.0, {}
        for item in self.__get_crossref_query_items(
            record=record, jour_vol_iss_list=jour_vol_iss_list, timeout=timeout
        ):
            try:
                retrieved_record_dict = connector_utils.json_to_record(item=item)
                similarity = self.__get_similarity(
                    record=record, retrieved_record_dict=retrieved_record_dict
                )
                retrieved_record = colrev.record.PrepRecord(data=retrieved_record_dict)

                # source = (
                #     f'https://api.crossref.org/works/{retrieved_record.data["doi"]}'
                # )
                # retrieved_record.add_provenance_all(source=source)

                # record.set_masterdata_complete(source=source)

                if jour_vol_iss_list:
                    record_list.append(retrieved_record)
                elif most_similar < similarity:
                    most_similar = similarity
                    most_similar_record = retrieved_record.get_data()
            except colrev_exceptions.RecordNotParsableException:
                pass

        if not jour_vol_iss_list:
            if most_similar_record:
                record_list = [colrev.record.PrepRecord(data=most_similar_record)]

        return record_list

    # def __check_journal(
    #     self,
    #     prep_operation: colrev.ops.prep.Prep,
    #     record: colrev.record.Record,
    #     timeout: int,
    #     save_feed: bool,
    # ) -> colrev.record.Record:
    #     """When there is no doi, journal names can be checked against crossref"""

    #     if record.data["ENTRYTYPE"] == "article":
    #         # If type article and doi not in record and
    #         # journal name not found in journal-query: notify
    #         journals = Journals(etiquette=self.etiquette)
    #         # record.data["journal"] = "Information Systems Research"
    #         found = False
    #         ret = journals.query(record.data["journal"])
    #         for rets in ret:
    #             if rets["title"]:
    #                 found = True
    #                 break
    #         if not found:
    #             record.add_masterdata_provenance_note(
    #                 key="journal", note="quality_defect:journal not in crossref"
    #             )

    #     return record

    def __get_masterdata_record(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        timeout: int,
        save_feed: bool,
    ) -> colrev.record.Record:
        try:
            if "doi" in record.data:
                retrieved_record = self.query_doi(
                    doi=record.data["doi"], etiquette=self.etiquette
                )
            else:
                retrieved_records = self.crossref_query(
                    record_input=record,
                    jour_vol_iss_list=False,
                    timeout=timeout,
                )
                retrieved_record = retrieved_records.pop()

                retries = 0
                while (
                    not retrieved_record
                    and retries < prep_operation.max_retries_on_error
                ):
                    retries += 1

                    retrieved_records = self.crossref_query(
                        record_input=record,
                        jour_vol_iss_list=False,
                        timeout=timeout,
                    )
                    retrieved_record = retrieved_records.pop()

            if 0 == len(retrieved_record.data) or "doi" not in retrieved_record.data:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="Record not found in crossref"
                )

            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=record, retrieved_record_original=retrieved_record
            )
            # prep_operation.review_manager.logger.debug("Found matching record")
            # prep_operation.review_manager.logger.debug(
            #     f"crossref similarity: {similarity} "
            #     f"(>{prep_operation.retrieval_similarity})"
            # )
            self.review_manager.logger.debug(
                f"crossref similarity: {similarity} "
                f"(<{prep_operation.retrieval_similarity})"
            )
            if similarity < prep_operation.retrieval_similarity:
                return record

            try:
                self.crossref_lock.acquire(timeout=120)

                # Note : need to reload file because the object is not shared between processes
                crossref_feed = self.search_source.get_feed(
                    review_manager=self.review_manager,
                    source_identifier=self.source_identifier,
                    update_only=False,
                )

                crossref_feed.set_id(record_dict=retrieved_record.data)
                crossref_feed.add_record(record=retrieved_record)

                record.merge(
                    merging_record=retrieved_record,
                    default_source=retrieved_record.data["colrev_origin"][0],
                )

                self.__prep_crossref_record(
                    record=record,
                    crossref_source=retrieved_record.data["colrev_origin"][0],
                )

                if save_feed:
                    crossref_feed.save_feed_file()

            except (
                colrev_exceptions.InvalidMerge,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                pass
            finally:
                try:
                    self.crossref_lock.release()
                except ValueError:
                    pass

            return record

        except (
            requests.exceptions.RequestException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ) as exc:
            if prep_operation.review_manager.verbose_mode:
                print(exc)

        return record

    def __check_doi_masterdata(
        self, record: colrev.record.Record
    ) -> colrev.record.Record:
        try:
            retrieved_record = self.query_doi(
                doi=record.data["doi"], etiquette=self.etiquette
            )
            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=record,
                retrieved_record_original=retrieved_record,
                same_record_type_required=False,
            )
            if similarity < 0.7:
                # self.review_manager.logger.error(
                #     f" mismatching metadata (record/doi-record): {record.data['doi']} "
                #     + f"(similarity: {similarity})"
                # )
                record.remove_field(key="doi")
                # record.print_citation_format()
                # retrieved_record.print_citation_format()

        except (
            requests.exceptions.RequestException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass

        return record

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 30,
    ) -> colrev.record.Record:
        """Retrieve masterdata from Crossref based on similarity with the record provided"""

        # To test the metadata provided for a particular DOI use:
        # https://api.crossref.org/works/DOI

        # https://github.com/OpenAPC/openapc-de/blob/master/python/import_dois.py
        if len(record.data.get("title", "")) < 35 and "doi" not in record.data:
            return record

        if "doi" in record.data:
            record = self.__check_doi_masterdata(record=record)

        record = self.__get_masterdata_record(
            prep_operation=prep_operation,
            record=record,
            timeout=timeout,
            save_feed=save_feed,
        )

        # Note: this should be optional
        # if "doi" not in record.data:
        #     record = self.__check_journal(
        #         prep_operation=prep_operation,
        #         record=record,
        #         timeout=timeout,
        #         save_feed=save_feed,
        #     )

        return record

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if source.filename.name != self.__crossref_md_filename.name:
            if not any(x in source.search_parameters for x in ["query", "scope"]):
                raise colrev_exceptions.InvalidQueryException(
                    "Crossref search_parameters requires a query or issn field"
                )

            if "scope" in source.search_parameters:
                if "issn" in source.search_parameters["scope"]:
                    assert isinstance(source.search_parameters["scope"]["issn"], list)
                    for issn_field in source.search_parameters["scope"]["issn"]:
                        if not re.match(self.__ISSN_REGEX, issn_field):
                            raise colrev_exceptions.InvalidQueryException(
                                f"Crossref journal issn ({issn_field}) not matching required format"
                            )
                elif "years" in source.search_parameters["scope"]:
                    years_field = source.search_parameters["scope"]["years"]
                    if not re.match(self.__YEAR_SCOPE_REGEX, years_field):
                        raise colrev_exceptions.InvalidQueryException(
                            f"Scope (years) ({years_field}) not matching required format"
                        )
                else:
                    raise colrev_exceptions.InvalidQueryException(
                        "Query missing valid parameters"
                    )

            elif "query" in source.search_parameters:
                # Note: not yet implemented/supported
                if " AND " in source.search_parameters["query"]:
                    raise colrev_exceptions.InvalidQueryException(
                        "AND not supported in CROSSREF query"
                    )

            else:
                raise colrev_exceptions.InvalidQueryException(
                    "Query missing valid parameters"
                )

            if source.search_type not in [
                colrev.settings.SearchType.DB,
                colrev.settings.SearchType.TOC,
            ]:
                raise colrev_exceptions.InvalidQueryException(
                    "Crossref search_type should be in [DB,TOC]"
                )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def __get_crossref_query_return(self, *, rerun: bool) -> typing.Iterator[dict]:
        params = self.search_source.search_parameters

        if "query" in params and "mode" not in params:
            crossref_query = {"bibliographic": params["query"].replace(" ", "+")}
            # potential extension : add the container_title:
            # crossref_query_return = works.query(
            #     container_title=
            #       "Journal of the Association for Information Systems"
            # )
            yield from self.__query(**crossref_query)
        elif "scope" in params and "issn" in params["scope"]:
            if "issn" in params["scope"]:
                for issn in params["scope"]["issn"]:
                    yield from self.__query_journal(issn=issn, rerun=rerun)

    def __restore_url(
        self,
        *,
        record: colrev.record.Record,
        feed: colrev.ops.search_feed.GeneralOriginFeed,
    ) -> None:
        """Restore the url from the feed if it exists
        (url-resolution is not always available)"""
        if record.data["ID"] not in feed.feed_records:
            return
        prev_url = feed.feed_records[record.data["ID"]].get("url", None)
        if prev_url is None:
            return
        record.data["url"] = prev_url

    def __run_md_search_update(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        crossref_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        records = search_operation.review_manager.dataset.load_records_dict()

        for feed_record_dict in crossref_feed.feed_records.values():
            feed_record = colrev.record.Record(data=feed_record_dict)

            try:
                retrieved_record = self.query_doi(
                    doi=feed_record_dict["doi"], etiquette=self.etiquette
                )

                if retrieved_record.data["doi"] != feed_record.data["doi"]:
                    continue

                crossref_feed.set_id(record_dict=retrieved_record.data)
            except (
                colrev_exceptions.RecordNotFoundInPrepSourceException,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                continue

            self.__prep_crossref_record(record=retrieved_record, prep_main_record=False)

            prev_record_dict_version = {}
            if retrieved_record.data["ID"] in crossref_feed.feed_records:
                prev_record_dict_version = crossref_feed.feed_records[
                    retrieved_record.data["ID"]
                ]

            self.__restore_url(record=retrieved_record, feed=crossref_feed)
            crossref_feed.add_record(record=retrieved_record)

            crossref_feed.update_existing_record(
                records=records,
                record_dict=retrieved_record.data,
                prev_record_dict_version=prev_record_dict_version,
                source=self.search_source,
                update_time_variant_fields=rerun,
            )

        crossref_feed.print_post_run_search_infos(records=records)
        crossref_feed.save_feed_file()
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

    def __scope_excluded(self, *, retrieved_record_dict: dict) -> bool:
        if (
            "scope" not in self.search_source.search_parameters
            or "years" not in self.search_source.search_parameters["scope"]
        ):
            return False
        year_from, year_to = self.search_source.search_parameters["scope"][
            "years"
        ].split("-")
        if not retrieved_record_dict.get("year", -1000).isdigit():
            return True
        if (
            int(year_from)
            <= int(retrieved_record_dict.get("year", -1000))
            <= int(year_to)
        ):
            return False
        return True

    def __run_keyword_exploration_search(
        self,
        search_operation: colrev.ops.search.Search,
        crossref_feed: colrev.ops.search_feed.GeneralOriginFeed,
    ) -> None:
        works = Works(etiquette=self.etiquette)

        def retrieve_exploratory_papers(keyword: str) -> typing.Iterator[dict]:
            crossref_query_return = works.query(bibliographic=keyword.replace(" ", "+"))
            yield from crossref_query_return

        records = search_operation.review_manager.dataset.load_records_dict()
        available_dois = [x["doi"] for x in records.values() if "doi" in x]

        covered_keywords = [
            x["explored_keyword"] for x in crossref_feed.feed_records.values()
        ]

        for keyword in self.search_source.search_parameters["query"].split(" OR "):
            self.review_manager.logger.info(f"Explore '{keyword}'")
            # Skip keywords that were already explored
            if keyword in covered_keywords:
                continue
            nr_added = 0
            for item in retrieve_exploratory_papers(keyword=keyword):
                try:
                    retrieved_record_dict = connector_utils.json_to_record(item=item)

                    # Skip papers that do not have the keyword in the title
                    if keyword not in retrieved_record_dict.get(
                        "title", ""
                    ).lower().replace("-", " "):
                        continue

                    # Skip papers that were already retrieved
                    if retrieved_record_dict["doi"] in available_dois:
                        continue
                    retrieved_record_dict["explored_keyword"] = keyword
                    crossref_feed.set_id(record_dict=retrieved_record_dict)
                    retrieved_record = colrev.record.Record(data=retrieved_record_dict)
                    self.__prep_crossref_record(
                        record=retrieved_record, prep_main_record=False
                    )

                    self.__restore_url(record=retrieved_record, feed=crossref_feed)

                    added = crossref_feed.add_record(record=retrieved_record)

                    if added:
                        nr_added += 1
                        search_operation.review_manager.logger.info(
                            " retrieve " + retrieved_record.data["doi"]
                        )
                    if nr_added >= 10:
                        break

                except (
                    colrev_exceptions.RecordNotParsableException,
                    colrev_exceptions.NotFeedIdentifiableException,
                    KeyError  # error in crossref package:
                    # if len(result['message']['items']) == 0:
                    # KeyError: 'items'
                ):
                    pass
                if nr_added < 10:
                    self.review_manager.logger.info(
                        f"Only {nr_added} papers found to resample keyword '{keyword}'"
                    )

        crossref_feed.print_post_run_search_infos(records=records)

        crossref_feed.save_feed_file()
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

        self.review_manager.dataset.format_records_file()
        self.review_manager.dataset.add_record_changes()
        self.review_manager.dataset.add_changes(path=self.search_source.filename)
        self.review_manager.create_commit(msg="Run search")

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        crossref_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        if rerun:
            search_operation.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        if self.search_source.search_parameters.get("mode", "") == "resample_keywords":
            self.__run_keyword_exploration_search(
                search_operation=search_operation, crossref_feed=crossref_feed
            )
            return

        records = search_operation.review_manager.dataset.load_records_dict()
        for item in self.__get_crossref_query_return(rerun=rerun):
            try:
                retrieved_record_dict = connector_utils.json_to_record(item=item)
                crossref_feed.set_id(record_dict=retrieved_record_dict)
                prev_record_dict_version = {}
                if retrieved_record_dict["ID"] in crossref_feed.feed_records:
                    prev_record_dict_version = deepcopy(
                        crossref_feed.feed_records[retrieved_record_dict["ID"]]
                    )

                if self.__scope_excluded(retrieved_record_dict=retrieved_record_dict):
                    continue
                retrieved_record = colrev.record.Record(data=retrieved_record_dict)
                self.__prep_crossref_record(
                    record=retrieved_record, prep_main_record=False
                )

                self.__restore_url(record=retrieved_record, feed=crossref_feed)

                added = crossref_feed.add_record(record=retrieved_record)

                if added:
                    search_operation.review_manager.logger.info(
                        " retrieve " + retrieved_record.data["doi"]
                    )
                else:
                    crossref_feed.update_existing_record(
                        records=records,
                        record_dict=retrieved_record.data,
                        prev_record_dict_version=prev_record_dict_version,
                        source=self.search_source,
                        update_time_variant_fields=rerun,
                    )

                # Note : only retrieve/update the latest deposits (unless in rerun mode)
                if not added and not rerun:
                    # problem: some publishers don't necessarily
                    # deposit papers chronologically
                    break
            except (
                colrev_exceptions.RecordNotParsableException,
                colrev_exceptions.NotFeedIdentifiableException,
                KeyError  # error in crossref package:
                # if len(result['message']['items']) == 0:
                # KeyError: 'items'
            ):
                pass

        crossref_feed.print_post_run_search_infos(records=records)

        crossref_feed.save_feed_file()
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Crossref"""

        self.__validate_source()

        crossref_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        try:
            if (
                self.search_source.is_md_source()
                or self.search_source.is_quasi_md_source()
            ):
                self.__run_md_search_update(
                    search_operation=search_operation,
                    crossref_feed=crossref_feed,
                    rerun=rerun,
                )

            else:
                self.__run_parameter_search(
                    search_operation=search_operation,
                    crossref_feed=crossref_feed,
                    rerun=rerun,
                )
        except (requests.exceptions.RequestException,) as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    self.__availability_exception_message
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                self.__availability_exception_message
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Crossref"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint"""

        if params and "https://search.crossref.org/?q=" in params:
            params = (
                params.replace("https://search.crossref.org/?q=", "")
                .replace("&from_ui=yes", "")
                .lstrip("+")
            )

            filename = operation.get_unique_filename(
                file_path_string=f"crossref_{params}"
            )
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.crossref",
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"query": params},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        if params is not None:
            params_dict: typing.Dict[str, typing.Any] = {"scope": {}}
            for item in params.split(";"):
                key, value = item.split("=")
                if key in ["issn", "years"]:
                    params_dict["scope"][key] = value
                else:
                    params_dict[key] = value
            if not params_dict["scope"]:
                del params_dict["scope"]

            if "scope" not in params_dict and "query" not in params_dict:
                raise colrev_exceptions.InvalidQueryException(
                    "Query parameters must contain query or scope (such as issn)"
                )

            filename = operation.get_unique_filename(
                file_path_string=f"crossref_{params}"
            )
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.crossref",
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters=params_dict,
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return
        cls.__add_interactively(operation=operation)

    @classmethod
    def __add_interactively(cls, *, operation: colrev.ops.search.Search) -> None:
        print("Interactively add Crossref as a SearchSource")
        print()
        print("Documentation:")
        print(
            "https://github.com/CoLRev-Environment/colrev/blob/"
            + "main/colrev/ops/built_in/search_sources/crossref.md"
        )
        print()
        query_type = ""
        while query_type not in ["j", "k"]:
            query_type = input("Create a query based on [k]eywords or [j]ournal?")
        if query_type == "j":
            print("Get ISSN from https://portal.issn.org/issn/search")

            issn = ""
            while True:
                issn = input("Enter the ISSN (or journal name to lookup the ISSN):")
                if re.match(cls.__ISSN_REGEX, issn):
                    break

                journals = Journals()
                ret = journals.query(issn)
                for jour in ret:
                    print(f"{jour['title']}: {','.join(jour['ISSN'])}")

            filename = operation.get_unique_filename(
                file_path_string=f"crossref_issn_{issn}"
            )
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.crossref",
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"scope": {"issn": [issn]}},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        # if query_type == "k":
        keywords = input("Enter the keywords:")

        filename = operation.get_unique_filename(
            file_path_string=f"crossref_{keywords}"
        )
        add_source = colrev.settings.SearchSource(
            endpoint="colrev.crossref",
            filename=filename,
            search_type=colrev.settings.SearchType.DB,
            search_parameters={"query": keywords},
            comment="",
        )
        operation.review_manager.settings.sources.append(add_source)

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Crossref"""

        source_item = [
            x
            for x in record.data["colrev_origin"]
            if str(source.filename).replace("data/search/", "") in x
        ]
        if source_item:
            record.set_masterdata_complete(
                source=source_item[0],
                masterdata_repository=self.review_manager.settings.is_curated_repo(),
            )

        return record



#! /usr/bin/env python
"""Checker for inconsistent-with-url-metadata."""
from __future__ import annotations

from thefuzz import fuzz

import colrev.ops.built_in.search_sources.website as website_connector
import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class InconsistentWithURLMetadataChecker:
    """The InconsistentWithURLMetadataChecker"""

    msg = "inconsistent-with-url-metadata"
    __fields_to_check = ["author", "title", "journal", "year", "volume", "number"]

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

        self.__url_connector = website_connector.WebsiteConnector(
            review_manager=quality_model.review_manager
        )

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the inconsistent-with-url-metadata checks"""

        if "url" not in record.data:
            return
        if any(x in record.data["url"] for x in ["search.ebscohost.com/login"]):
            return
        if "md_curated.bib" in record.data["colrev_data_provenance"]["url"]["source"]:
            return

        if self.__url_metadata_conflicts(record=record):
            record.add_masterdata_provenance_note(key="url", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="url", note=self.msg)

    def __url_metadata_conflicts(self, *, record: colrev.record.Record) -> bool:
        url_record = record.copy_prep_rec()
        self.__url_connector.retrieve_md_from_website(record=url_record)
        for key, value in url_record.data.items():
            if key not in self.__fields_to_check:
                continue
            if not isinstance(value, str):
                continue
            if key in record.data:
                if len(url_record.data[key]) < 5 or len(record.data[key]) < 5:
                    continue
                if (
                    fuzz.partial_ratio(
                        record.data[key].lower(), url_record.data[key].lower()
                    )
                    < 70
                ):
                    return True

        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentWithURLMetadataChecker(quality_model))



#! /usr/bin/env python
"""Screen based on CLI"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.screen.utils as util_cli_screen
import colrev.record
import colrev.settings
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.screen


@zope.interface.implementer(colrev.env.package_manager.ScreenPackageEndpointInterface)
@dataclass
class CoLRevCLIScreen(JsonSchemaMixin):
    """Screen documents using a CLI"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        self.__i = 0
        self.__pad = 0
        self.__stat_len = 0
        self.screening_criteria: dict = {}
        self.criteria_available = 0

    def __print_screening_criteria(
        self, *, screen_operation: colrev.ops.screen.Screen
    ) -> None:
        if not screen_operation.review_manager.settings.screen.criteria:
            return
        print("\nIn the screen, the following criteria are applied:\n")
        for (
            criterion_name,
            criterion_settings,
        ) in screen_operation.review_manager.settings.screen.criteria.items():
            color = colors.GREEN
            if (
                colrev.settings.ScreenCriterionType.exclusion_criterion
                == criterion_settings.criterion_type
            ):
                color = colors.RED
            print(
                f" - {criterion_name} "
                f"({color}{criterion_settings.criterion_type}{colors.END}): "
                f"{criterion_settings.explanation}"
            )
            if criterion_settings.comment != "":
                print(f"   {criterion_settings.comment}")

    def __screen_with_criteria_print_overall_decision(
        self, *, record: colrev.record.Record, screen_inclusion: bool
    ) -> None:
        if screen_inclusion:
            print(
                f"Overall screening decision for {record.data['ID']}: "
                f"{colors.GREEN}include{colors.END}"
            )
        else:
            print(
                f"Overall screening decision for {record.data['ID']}: "
                f"{colors.RED}exclude{colors.END}"
            )

    def __screen_record_with_criteria(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        record: colrev.record.Record,
        abstract_from_tei: bool,
    ) -> str:
        decisions = []
        quit_pressed, skip_pressed = False, False

        for criterion_name, criterion_settings in self.screening_criteria.items():
            decision, ret = "NA", "NA"
            while ret not in ["y", "n", "q", "s"]:
                color = colors.GREEN
                if (
                    colrev.settings.ScreenCriterionType.exclusion_criterion
                    == criterion_settings.criterion_type
                ):
                    color = colors.RED

                ret = input(
                    # is relevant / should be in the sample / should be retained
                    # ({self.__i}/{self.__stat_len})
                    f"Record should be included according to"
                    f" {criterion_settings.criterion_type}"
                    f" {color}{criterion_name}{colors.END}"
                    " [y,n,q,s for yes,no,quit,skip to decide later]? "
                )
                if ret == "q":
                    quit_pressed = True
                elif ret == "s":
                    skip_pressed = True
                    continue
                elif ret in ["y", "n"]:
                    decision = ret

            if quit_pressed:
                return "quit"
            if skip_pressed:
                return "skip"

            decision = decision.replace("n", "out").replace("y", "in")
            decisions.append([criterion_name, decision])

        c_field = ""
        for criterion_name, decision in decisions:
            c_field += f";{criterion_name}={decision}"
        c_field = c_field.replace(" ", "").lstrip(";")

        screen_inclusion = all(decision == "in" for _, decision in decisions)
        self.__screen_with_criteria_print_overall_decision(
            record=record, screen_inclusion=screen_inclusion
        )

        if abstract_from_tei:
            if "abstract" in record.data:
                del record.data["abstract"]

        screen_operation.screen(
            record=record,
            screen_inclusion=screen_inclusion,
            screening_criteria=c_field,
            PAD=self.__pad,
        )
        return "screened"

    def __screen_record_without_criteria(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        record: colrev.record.Record,
        abstract_from_tei: bool,
    ) -> str:
        quit_pressed = False
        decision, ret = "NA", "NA"
        while ret not in ["y", "n", "q", "s"]:
            ret = input(
                f"({self.__i}/{self.__stat_len}) "
                "Include [y,n,q,s for yes, no, quit, skip/screen later]? "
            )
            if ret == "q":
                quit_pressed = True
            elif ret == "s":
                return "skip"
            elif ret in ["y", "n"]:
                decision = ret

        if quit_pressed:
            screen_operation.review_manager.logger.info("Stop screen")
            return "quit"

        if abstract_from_tei:
            if "abstract" in record.data:
                del record.data["abstract"]
        if decision == "y":
            screen_operation.screen(
                record=record,
                screen_inclusion=True,
                screening_criteria="NA",
            )
        if decision == "n":
            screen_operation.screen(
                record=record,
                screen_inclusion=False,
                screening_criteria="NA",
                PAD=self.__pad,
            )
        return "screened"

    def __screen_record(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        record_dict: dict,
    ) -> str:
        record = colrev.record.Record(data=record_dict)
        abstract_from_tei = False
        if (
            "abstract" not in record.data
            and Path(record.data.get("file", "")).suffix == ".pdf"
        ):
            try:
                abstract_from_tei = True
                tei = screen_operation.review_manager.get_tei(
                    pdf_path=Path(record.data["file"]),
                    tei_path=record.get_tei_filename(),
                )
                record.data["abstract"] = tei.get_abstract()
            except colrev_exceptions.TEIException:
                pass

        self.__i += 1
        print("\n\n")
        print(f"Record {self.__i} (of {self.__stat_len})")
        print(record)

        if self.criteria_available:
            ret = self.__screen_record_with_criteria(
                screen_operation=screen_operation,
                record=record,
                abstract_from_tei=abstract_from_tei,
            )

        else:
            ret = self.__screen_record_without_criteria(
                screen_operation=screen_operation,
                record=record,
                abstract_from_tei=abstract_from_tei,
            )

        return ret

    def __screen_cli(
        self, screen_operation: colrev.ops.screen.Screen, split: list
    ) -> dict:
        screen_data = screen_operation.get_data()
        self.__pad = screen_data["PAD"]
        self.__stat_len = screen_data["nr_tasks"]

        if 0 == self.__stat_len:
            screen_operation.review_manager.logger.info("No records to prescreen")

        records = screen_operation.review_manager.dataset.load_records_dict()

        self.__print_screening_criteria(screen_operation=screen_operation)

        self.screening_criteria = (
            util_cli_screen.get_screening_criteria_from_user_input(
                screen_operation=screen_operation, records=records
            )
        )
        self.criteria_available = len(self.screening_criteria.keys())

        for record_dict in screen_data["items"]:
            if record_dict["ID"] not in split:
                continue

            ret = self.__screen_record(
                screen_operation=screen_operation, record_dict=record_dict
            )

            if ret == "skip":
                continue
            if ret == "quit":
                screen_operation.review_manager.logger.info("Stop screen")
                break

        if self.__stat_len == 0:
            screen_operation.review_manager.logger.info("No records to screen")
            return records

        screen_operation.review_manager.dataset.add_record_changes()

        if self.__i < self.__stat_len and split:  # if records remain for screening
            if input("Create commit (y/n)?") != "y":
                return records

        screen_operation.review_manager.create_commit(
            msg="Screening (manual)", manual_author=True
        )
        return records

    def run_screen(
        self, screen_operation: colrev.ops.screen.Screen, records: dict, split: list
    ) -> dict:
        """Screen records based on a cli"""

        records = self.__screen_cli(screen_operation, split)

        return records



#! /usr/bin/env python
"""Template for a custom Prescreen PackageEndpoint"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING

import zope.interface
from dacite import from_dict

import colrev.operation
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prescreen

# pylint: disable=too-few-public-methods


@zope.interface.implementer(
    colrev.env.package_manager.PrescreenPackageEndpointInterface
)
class CustomPrescreen:
    """Class for custom prescreen scripts"""

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = from_dict(
            data_class=colrev.env.package_manager.DefaultSettings, data=settings
        )

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen the record"""

        for record in records.values():
            if random.random() < 0.5:  # nosec
                prescreen_operation.prescreen(
                    record=colrev.record.Record(data=record), prescreen_inclusion=True
                )

            else:
                prescreen_operation.prescreen(
                    record=colrev.record.Record(data=record), prescreen_inclusion=False
                )

        prescreen_operation.review_manager.dataset.save_records_dict(records=records)
        prescreen_operation.review_manager.dataset.add_record_changes()
        prescreen_operation.review_manager.create_commit(
            msg="Pre-screen (random)",
            manual_author=False,
            script_call="colrev prescreen",
        )

        return records



#! /usr/bin/env python
"""Checker for thesis-with-multiple-authors."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class ThesisWithMultipleAuthorsChecker:
    """The ThesisWithMultipleAuthorsChecker"""

    msg = "thesis-with-multiple-authors"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the thesis-with-multiple-authors checks"""

        if self.__multiple_authored_thesis(record=record):
            record.add_masterdata_provenance_note(key="author", note=self.msg)
        else:
            record.remove_masterdata_provenance_note(key="author", note=self.msg)

    def __multiple_authored_thesis(self, *, record: colrev.record.Record) -> bool:
        if record.data["ENTRYTYPE"] in [
            "thesis",
            "phdthesis",
            "mastertsthesis",
        ] and " and " in record.data.get("author", ""):
            return True
        return False


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ThesisWithMultipleAuthorsChecker(quality_model))



#! /usr/bin/env python
"""Retrieval of PDFs from the LocalIndex"""
from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.pdf_get

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PDFGetPackageEndpointInterface)
@dataclass
class LocalIndexPDFGet(JsonSchemaMixin):
    """Get PDFs from LocalIndex"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def get_pdf(
        self, pdf_get_operation: colrev.ops.pdf_get.PDFGet, record: colrev.record.Record
    ) -> colrev.record.Record:
        """Get PDFs from the local-index"""

        local_index = pdf_get_operation.review_manager.get_local_index()

        try:
            retrieved_record = local_index.retrieve(
                record_dict=record.data, include_file=True
            )
        except colrev_exceptions.RecordNotInIndexException:
            return record

        if "file" in retrieved_record:
            record.update_field(
                key="file", value=str(retrieved_record["file"]), source="local_index"
            )
            pdf_get_operation.import_pdf(record=record)
            if "fulltext" in retrieved_record:
                try:
                    record.get_tei_filename().write_text(retrieved_record["fulltext"])
                except FileNotFoundError:
                    pass
                del retrieved_record["fulltext"]
            else:
                tei_ext_path = Path(
                    retrieved_record["file"]
                    .replace("pdfs/", ".tei/")
                    .replace(".pdf", ".tei.xml")
                )
                if tei_ext_path.is_file():
                    new_path = record.get_tei_filename()
                    new_path.resolve().parent.mkdir(exist_ok=True, parents=True)
                    shutil.copy(tei_ext_path, new_path)

        return record



#! /usr/bin/env python
"""SearchSource: AIS electronic Library"""
from __future__ import annotations

import json
import re
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urlparse

import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_enl
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class AISeLibrarySearchSource(JsonSchemaMixin):
    """SearchSource for the AIS electronic Library (AISeL)"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    endpoint = "colrev.ais_library"
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "AIS eLibrary"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/aisel.md"
    )

    __conference_abbreviations = {
        "ICIS": "International Conference on Information Systems",
        "PACIS": "Pacific-Asia Conference on Information Systems",
        "ECIS": "European Conference on Information Systems",
        "AMCIS": "Americas Conference on Information Systems",
        "HICSS": "Hawaii International Conference on System Sciences",
        "MCIS": "Mediterranean Conference on Information Systems",
        "ACIS": "Australasian Conference on Information Systems",
    }

    __link_confs = {
        "https://aisel.aisnet.org/hicss": "Hawaii International Conference on System Sciences",
        "https://aisel.aisnet.org/amcis": "Americas Conference on Information Systems",
        "https://aisel.aisnet.org/pacis": "Pacific-Asia Conference on Information Systems",
        "https://aisel.aisnet.org/ecis": "European Conference on Information Systems",
        "https://aisel.aisnet.org/icis": "International Conference on Information Systems",
    }

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.review_manager = source_operation.review_manager
        self.quality_model = self.review_manager.get_qm()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for AIS electronic Library (AISeL)"""

        result = {"confidence": 0.0}
        # TBD: aisel does not return bibtex?!
        nr_ais_links = data.count("https://aisel.aisnet.org/")
        nr_items = data.count("\n@")
        if nr_items > 0 and "colrev_status" not in data:
            result["confidence"] = nr_ais_links / nr_items

        if "%T " in data and "colrev_status" not in data:
            if data.count("%U https://aisel.aisnet.org") > 0.9 * data.count("%T "):
                result["confidence"] = 1.0

        if "@article" in data or "@inproc" in data and "colrev_status" not in data:
            if data.count("https://aisel.aisnet.org") > 0.9 * data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def __parse_query(cls, *, query: str) -> dict:
        peer_reviewed = "peer_reviewed=true" in query
        start_date = ""
        if "start_date=" in query:
            start_date = query[query.find("start_date=") + 11 :]
            start_date = start_date[: start_date.find("&")]
            start_date = start_date.replace("%2F", "/")
        end_date = ""
        if "end_date=" in query:
            end_date = query[query.find("end_date=") + 9 :]
            end_date = end_date[: end_date.find("&")]
            end_date = end_date.replace("%2F", "/")

        query = query[query.find("?q=") + 3 : query.find("&start")]
        query_parts = urllib.parse.unquote(query).split(" ")

        search_terms = []
        query_parts_merged = []
        parenthesis_expression = ""
        for query_part in query_parts:
            if query_part not in ["(", ")"] and "" == parenthesis_expression:
                query_parts_merged.append(query_part)
            elif query_part == "(":
                parenthesis_expression += "("
            elif query_part == ")":
                parenthesis_expression = parenthesis_expression.rstrip().replace(
                    "(", ""
                )
                query_parts_merged.append(parenthesis_expression)
                parenthesis_expression = ""
            else:
                parenthesis_expression = parenthesis_expression + query_part + " "

        term_no = 1
        operator = ""

        for query_part in query_parts_merged:
            if query_part in ["OR", "AND", "NOT"]:
                operator = query_part
                continue

            field = "All fields"
            if "%3A" in query_part:
                field, query_part = query_part.split("%3A")
            search_term = {"operator": operator, "term": query_part, "field": field}

            search_terms.append(search_term)
            term_no += 1

        params = {"search_terms": search_terms, "peer_reviewed": peer_reviewed}
        if start_date:
            params["start_date"] = start_date
        if end_date:
            params["end_date"] = end_date
        return params

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        # if params is None:
        #     operation.add_interactively(endpoint=cls.endpoint)
        #     return

        params = params.lstrip("colrev.ais_library:").rstrip('"').lstrip('"')

        host = urlparse(params).hostname

        if host and host.endswith("aisel.aisnet.org"):
            q_params = cls.__parse_query(query=params)
            filename = operation.get_unique_filename(file_path_string="ais")
            add_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"query": q_params},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise colrev_exceptions.PackageParameterError(
            f"Cannot add aisel endpoint with query {q_params}"
        )

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if "query" in source.search_parameters:
            if "search_terms" not in source.search_parameters["query"]:
                raise colrev_exceptions.InvalidQueryException(
                    "query parameter does not contain search_terms"
                )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def __get_ais_query_return(self) -> list:
        def query_from_params(params: dict) -> str:
            final = ""
            for i in params["search_terms"]:
                final = f"{final} {i['operator']} {i['field']}:{i['term']}".strip()
                final = final.replace("All fields:", "")

            if "start_date" in params["search_terms"]:
                final = f"{final}&start_date={params['search_terms']['start_date']}"

            if "end_date" in params["search_terms"]:
                final = f"{final}&end_date={params['search_terms']['end_date']}"

            if "peer_reviewed" in params["search_terms"]:
                final = f"{final}&peer_reviewed=true"

            return urllib.parse.quote(final)

        params = self.search_source.search_parameters["query"]
        final_q = query_from_params(params)

        query_string = (
            "https://aisel.aisnet.org/do/search/results/refer?"
            + "start=0&context=509156&sort=score&facet=&dlt=Export122204"
        )
        query_string = f"{query_string}&q={final_q}"

        response = requests.get(query_string, timeout=300)
        response.raise_for_status()

        zotero_translation_service = (
            self.review_manager.get_zotero_translation_service()
        )
        zotero_translation_service.start()

        headers = {"Content-type": "text/plain"}
        ret = requests.post(
            "http://127.0.0.1:1969/import",
            headers=headers,
            data=response.content,
            timeout=30,
        )
        headers = {"Content-type": "application/json"}

        try:
            json_content = json.loads(ret.content)
            export = requests.post(
                "http://127.0.0.1:1969/export?format=bibtex",
                headers=headers,
                json=json_content,
                timeout=30,
            )

        except Exception as exc:
            raise colrev_exceptions.ImportException(
                f"Zotero translators failed ({exc})"
            )

        records = self.review_manager.dataset.load_records_dict(
            load_str=export.content.decode("utf-8")
        )

        return list(records.values())

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        ais_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        # pylint: disable=too-many-branches

        if rerun:
            search_operation.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        records = search_operation.review_manager.dataset.load_records_dict()
        for record_dict in self.__get_ais_query_return():
            # Note : discard "empty" records
            if "" == record_dict.get("author", "") and "" == record_dict.get(
                "title", ""
            ):
                continue

            try:
                ais_feed.set_id(record_dict=record_dict)
            except colrev_exceptions.NotFeedIdentifiableException:
                continue

            # prev_record_dict_version = {}
            # if record_dict["ID"] in ais_feed.feed_records:
            #     prev_record_dict_version = deepcopy(
            #         ais_feed.feed_records[record_dict["ID"]]
            #     )

            prep_record = colrev.record.PrepRecord(data=record_dict)

            if "colrev_data_provenance" in prep_record.data:
                del prep_record.data["colrev_data_provenance"]

            added = ais_feed.add_record(record=prep_record)

            if added:
                self.review_manager.logger.info(" retrieve " + prep_record.data["url"])
            # else:
            #     search_operation.update_existing_record(
            #         records=records,
            #         record_dict=prep_record.data,
            #         prev_record_dict_version=prev_record_dict_version,
            #         source=self.search_source,
            #         update_time_variant_fields=rerun,
            #     )

        ais_feed.print_post_run_search_infos(records=records)

        ais_feed.save_feed_file()
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of AISeLibrary"""

        self.__validate_source()

        ais_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        try:
            #  Note: not (yet) supported: self.search_source.is_md_source()
            if "query" in self.search_source.search_parameters:
                self.__run_parameter_search(
                    search_operation=search_operation,
                    ais_feed=ais_feed,
                    rerun=rerun,
                )
        except (
            requests.exceptions.JSONDecodeError,
            requests.exceptions.HTTPError,
        ) as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Crossref (check https://status.crossref.org/)"
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref (check https://status.crossref.org/) ({exc})"
            )

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix in [".txt", ".enl"]:
            enl_loader = colrev.ops.load_utils_enl.ENLLoader(
                load_operation=load_operation, source=self.search_source
            )
            records = enl_loader.load(source=self.search_source)
            return records

        # for API-based searches
        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def __fix_entrytype(self, *, record: colrev.record.Record) -> None:
        # Note : simple heuristic
        # but at the moment, AISeLibrary only indexes articles and conference papers
        if (
            record.data.get("volume", "UNKNOWN") != "UNKNOWN"
            or record.data.get("number", "UNKNOWN") != "UNKNOWN"
        ) and not any(
            x in record.data.get("journal", "")
            for x in [
                "HICSS",
                "ICIS",
                "ECIS",
                "AMCIS",
                "Proceedings",
                "All Sprouts Content",
            ]
        ):
            # Journal articles
            if (
                "journal" not in record.data
                and "title" in record.data
                and "chapter" in record.data
            ):
                record.rename_field(key="title", new_key="journal")
                record.rename_field(key="chapter", new_key="title")
                record.remove_field(key="publisher")

            record.change_entrytype(new_entrytype="article", qm=self.quality_model)
        else:
            # Inproceedings

            record.remove_field(key="publisher")

            if (
                "booktitle" not in record.data
                and "title" in record.data
                and "chapter" in record.data
            ):
                record.rename_field(key="title", new_key="booktitle")
                record.rename_field(key="chapter", new_key="title")

            record.change_entrytype(
                new_entrytype="inproceedings", qm=self.quality_model
            )

            if record.data.get("booktitle", "") in [
                "Research-in-Progress Papers",
                "Research Papers",
            ]:
                if "https://aisel.aisnet.org/ecis" in record.data.get("url", ""):
                    record.update_field(
                        key="booktitle", value="ECIS", source="prep_ais_source"
                    )

    def __unify_container_titles(self, *, record: colrev.record.Record) -> None:
        if "https://aisel.aisnet.org/misq/" in record.data.get("url", ""):
            record.update_field(
                key="journal", value="MIS Quarterly", source="prep_ais_source"
            )
            record.remove_field(key="booktitle")

        if "https://aisel.aisnet.org/misqe/" in record.data.get("url", ""):
            record.update_field(
                key="journal", value="MIS Quarterly Executive", source="prep_ais_source"
            )
            record.remove_field(key="booktitle")

        if "https://aisel.aisnet.org/bise/" in record.data.get("url", ""):
            record.update_field(
                key="journal",
                value="Business & Information Systems Engineering",
                source="prep_ais_source",
            )
            record.remove_field(key="booktitle")

        if record.data["ENTRYTYPE"] == "inproceedings":
            for conf_abbreviation, conf_name in self.__conference_abbreviations.items():
                if conf_abbreviation in record.data.get("booktitle", ""):
                    record.update_field(
                        key="booktitle",
                        value=conf_name,
                        source="prep_ais_source",
                    )

        for link_part, conf_name in self.__link_confs.items():
            if link_part in record.data.get("url", ""):
                record.update_field(
                    key="booktitle",
                    value=conf_name,
                    source="prep_ais_source",
                )

    def __format_fields(self, *, record: colrev.record.Record) -> None:
        if "abstract" in record.data:
            if record.data["abstract"] == "N/A":
                record.remove_field(key="abstract")
        if "author" in record.data:
            record.update_field(
                key="author",
                value=record.data["author"].replace("\n", " "),
                source="prep_ais_source",
                keep_source_if_equal=True,
            )

    def __exclude_complementary_material(self, *, record: colrev.record.Record) -> None:
        if re.match(
            r"MISQ Volume \d{1,2}, Issue \d Table of Contents",
            record.data.get("title", ""),
        ):
            record.prescreen_exclude(reason="complementary material")

    def prepare(
        self, record: colrev.record.PrepRecord, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for the AIS electronic Library (AISeL)"""

        self.__fix_entrytype(record=record)
        self.__unify_container_titles(record=record)
        self.__format_fields(record=record)
        self.__exclude_complementary_material(record=record)

        return record



#! /usr/bin/env python
"""Jupyter notebook for prep-man operation"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.env.utils
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep_man

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepManPackageEndpointInterface)
@dataclass
class CurationJupyterNotebookManPrep(JsonSchemaMixin):
    """Manual preparation based on a Jupyter Notebook"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self, *, prep_man_operation: colrev.ops.prep_man.PrepMan, settings: dict
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        Path("prep_man").mkdir(exist_ok=True)
        if not Path("prep_man/prep_man_curation.ipynb").is_file():
            prep_man_operation.review_manager.logger.info(
                f"Activated jupyter notebook to"
                f"{Path('prep_man/prep_man_curation.ipynb')}"
            )
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/ops/prep_man_curation.ipynb"),
                target=Path("prep_man/prep_man_curation.ipynb"),
            )

    def prepare_manual(
        self,
        prep_man_operation: colrev.ops.prep_man.PrepMan,  # pylint: disable=unused-argument
        records: dict,
    ) -> dict:
        """Prepare records manually based on  a Jupyter notebeook"""

        input(
            "Navigate to the jupyter notebook available at\n"
            "prep_man/prep_man_curation.ipynb\n"
            "Press Enter to continue."
        )
        return records



#! /usr/bin/env python
"""Loggers for CoLRev"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import colrev.exceptions as colrev_exceptions

if TYPE_CHECKING:
    import colrev.review_manager


def setup_logger(
    *, review_manager: colrev.review_manager.ReviewManager, level: int = logging.INFO
) -> logging.Logger:
    """Setup the CoLRev logger"""

    # for logger debugging:
    # from logging_tree import printout
    # printout()
    logger = logging.getLogger(f"colrev{str(review_manager.path).replace('/', '_')}")
    logger.setLevel(level)

    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)

    formatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(level)

    logger.addHandler(handler)
    logger.propagate = False

    return logger


def setup_report_logger(
    *, review_manager: colrev.review_manager.ReviewManager, level: int = logging.INFO
) -> logging.Logger:
    """Setup the report logger (used for git commit report)"""

    try:
        report_logger = logging.getLogger(
            f"colrev_report{str(review_manager.path).replace('/', '_')}"
        )

        if report_logger.handlers:
            for handler in report_logger.handlers:
                report_logger.removeHandler(handler)

        report_logger.setLevel(level)
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        report_file_handler = logging.FileHandler(review_manager.report_path, mode="a")
        report_file_handler.setFormatter(formatter)

        report_logger.addHandler(report_file_handler)

        if logging.DEBUG == level:
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            report_logger.addHandler(handler)
        report_logger.propagate = False
    except FileNotFoundError as exc:
        raise colrev_exceptions.RepoSetupError("Missing file") from exc
    return report_logger


def reset_report_logger(*, review_manager: colrev.review_manager.ReviewManager) -> None:
    """Reset the report log file (used for the git commit report)"""

    if review_manager.report_logger.handlers:
        report_handler = review_manager.report_logger.handlers[0]
        review_manager.report_logger.removeHandler(report_handler)
        report_handler.close()

    if review_manager.report_path.is_file():
        with open(review_manager.report_path, "r+", encoding="utf8") as file:
            file.truncate(0)

    file_handler = logging.FileHandler(review_manager.report_path, mode="a")
    file_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    file_handler.setFormatter(formatter)
    review_manager.report_logger.addHandler(file_handler)



#!/usr/bin/env python3
"""Scripts printing information for the colrev show command"""
import platform
from pathlib import Path

import colrev.record
import colrev.ui_cli.cli_colors as colors


def print_sample(review_manager: colrev.review_manager.ReviewManager) -> None:
    """Print the sample on cli"""

    colrev.operation.CheckOperation(review_manager=review_manager)
    records = review_manager.dataset.load_records_dict()
    sample = [
        r
        for r in records.values()
        if r["colrev_status"]
        in [
            colrev.record.RecordState.rev_synthesized,
            colrev.record.RecordState.rev_included,
        ]
    ]
    if 0 == len(sample):
        print("No records included in sample (yet)")

    for sample_r in sample:
        colrev.record.Record(data=sample_r).print_citation_format()


def print_venv_notes() -> None:
    """Print the virtual environment details on cli"""

    current_platform = platform.system()
    if current_platform == "Linux":
        print("Detected platform: Linux")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {colors.ORANGE}python3 -m venv venv{colors.END}")
        print("To activate virtualenv, run")
        print(f"  {colors.ORANGE}source venv/bin/activate{colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {colors.ORANGE}python -m pip install colrev{colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {colors.ORANGE}deactivate{colors.END}")
    elif current_platform == "Darwin":
        print("Detected platform: MacOS")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {colors.ORANGE}python3 -m venv venv{colors.END}")
        print("To activate virtualenv, run")
        print(f"  {colors.ORANGE}source venv/bin/activate{colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {colors.ORANGE}python -m pip install colrev{colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {colors.ORANGE}deactivate{colors.END}")
    elif current_platform == "Windows":
        print("Detected platform: Windows")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {colors.ORANGE}python -m venv venv{colors.END}")
        print("To activate virtualenv, run")
        print(f"  {colors.ORANGE}venv\\Scripts\\Activate.ps1{colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {colors.ORANGE}python -m pip install colrev{colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {colors.ORANGE}deactivate{colors.END}")
    else:
        print(
            "Platform not detected... "
            "cannot provide infos in how to activate virtualenv"
        )



#! /usr/bin/env python
"""Checker for record-not-in-toc."""
from __future__ import annotations

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class RecordNotInTOCChecker:
    """The RecordNotInTOCChecker"""

    msg = "record-not-in-toc"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model
        self.local_index = colrev.env.local_index.LocalIndex(verbose_mode=False)

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the record-not-in-toc checks"""

        try:
            # Search within the table-of-content in local_index
            self.local_index.retrieve_from_toc(
                record_dict=record.data,
                similarity_threshold=0.9,
                include_file=False,
            )
            if "journal" in record.data:
                record.remove_masterdata_provenance_note(key="journal", note=self.msg)
            elif "booktitle" in record.data:
                record.remove_masterdata_provenance_note(key="booktitle", note=self.msg)
        except colrev.exceptions.RecordNotInIndexException:
            pass
        except colrev_exceptions.RecordNotInTOCException:
            if "journal" in record.data:
                record.add_masterdata_provenance_note(key="journal", note=self.msg)
            elif "booktitle" in record.data:
                record.add_masterdata_provenance_note(key="booktitle", note=self.msg)


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(RecordNotInTOCChecker(quality_model))
